import warnings
import os
# Ensure worker processes inherit the warning filter; set before other imports.
os.environ['PYTHONWARNINGS'] = 'ignore'
import numpy as np
import pytest
from astropy import units as u
from astropy.io import fits
from astropy.modeling import models
from astropy.nddata import StdDevUncertainty
from astropy.tests.helper import assert_quantity_allclose
from astropy.wcs import WCS
from glue.core.roi import XRangeROI
from numpy.testing import assert_allclose, assert_array_equal
from specutils.spectra import Spectrum
from specutils import SpectralRegion
from traitlets import TraitError

from jdaviz.configs.default.plugins.model_fitting import fitting_backend as fb
from jdaviz.configs.default.plugins.model_fitting import initializers
from jdaviz.core.custom_units_and_equivs import PIX2
from jdaviz.conftest import _create_spectrum1d_cube_with_fluxunit

SPECTRUM_SIZE = 200  # length of spectrum


def build_spectrum(sigma=0.1):
    g1 = models.Gaussian1D(1, 4.6, 0.2)
    g2 = models.Gaussian1D(2.5, 5.5, 0.1)
    g3 = models.Gaussian1D(-1.7, 8.2, 0.1)

    x = np.linspace(0, 10, SPECTRUM_SIZE)
    y = g1(x) + g2(x) + g3(x)

    noise = np.random.normal(4., sigma, x.shape)

    return x, y + noise


def test_model_params():
    # expected model parameters:
    model_parameters = {"Gaussian1D": ["amplitude", "stddev", "mean"],
                        "Const1D": ["amplitude"],
                        "Linear1D": ["slope", "intercept"],
                        "Polynomial1D": ["c0", "c1"],
                        "PowerLaw1D": ["amplitude", "x_0", "alpha"],
                        "Lorentz1D": ["amplitude", "x_0", "fwhm"],
                        "Voigt1D": ["x_0", "amplitude_L", "fwhm_L", "fwhm_G"],
                        "BlackBody": ["temperature", "scale"],
                        "Spline1D": [],
                        }

    for model_name in initializers.MODELS.keys():
        if model_name not in model_parameters.keys():
            # this would be caught later by the assertion anyways,
            # but raising an error will be more clear that the
            # test needs to be updated rather than the code breaking
            raise ValueError(f"{model_name} not in test dictionary of expected parameters")
        expected_params = model_parameters.get(model_name, [])

        params = []
        # SplineSmoothingFitter does not have knot points, so no scalar parameters
        # Will need to account for if SplineExactKnotsFitter is implemented
        if model_name != "Spline1D":
            params = initializers.get_model_parameters(model_name)
        assert len(params) == len(expected_params)
        assert np.all([p in expected_params for p in params])


def test_check_poly_order_function(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)
    plugin = specviz_helper.plugins["Model Fitting"]._obj
    plugin.dataset_selected = '1D Spectrum'

    result = plugin._check_poly_order()
    assert result is None

    result = plugin._check_poly_order(model_comp='fake')
    assert result is None

    result = plugin._check_poly_order(poly_order=-1.5)
    assert plugin.poly_order_invalid_msg == ""
    assert result is None

    result = plugin._check_poly_order(model_comp='fake', poly_order=-1.5)
    assert plugin.poly_order_invalid_msg == ""
    assert result is None

    poly_order_default = 0
    assert plugin.poly_order == poly_order_default  # stays default

    plugin.poly_order = poly_order_default
    plugin.model_comp_selected = 'Polynomial1D'

    result = plugin._check_poly_order()
    assert result == poly_order_default

    result = plugin._check_poly_order(model_comp='Polynomial1D')
    assert result == poly_order_default

    result = plugin._check_poly_order(poly_order=4)
    assert result == 4
    # doesn't change actual value of plugin.poly_order
    assert plugin.poly_order == poly_order_default

    result = plugin._check_poly_order(model_comp='Polynomial1D', poly_order=5)
    assert result == 5
    # doesn't change actual value of plugin.poly_order
    assert plugin.poly_order == poly_order_default

    for poly_order in [-2, -1, 2.5]:
        result = plugin._check_poly_order(poly_order=poly_order)
        assert plugin.poly_order_invalid_msg == "Order must be an integer >= 0"
        assert result is None


def test_check_poly_order_observer(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)
    plugin = specviz_helper.plugins["Model Fitting"]._obj
    plugin.dataset_selected = '1D Spectrum'

    # poly_order default
    assert plugin.poly_order == 0
    # No model components added yet
    assert len(plugin.model_components) == 0

    # Make sure that the observer doesn't interfere with the other models
    non_poly_model_comps = [m for m in initializers.MODELS.keys() if m != "Polynomial1D"]
    for i, model_comp in enumerate(non_poly_model_comps):
        plugin.model_comp_selected = model_comp
        # Run the observer by setting poly_order
        plugin.poly_order -= 1
        plugin.vue_add_model({})
        # Check that the model component exists (i.e. was added successfully)
        assert len(plugin.model_components) == i + 1
        assert plugin.model_components[i]

    plugin.model_comp_selected = "Polynomial1D"

    vue_err_msg = "Order must be an integer >= 0"
    err_msg = f"poly_{vue_err_msg.lower()}"
    with pytest.raises(ValueError, match=err_msg):
        plugin.vue_add_model({})

    for poly_order in range(-3, 0):
        plugin.poly_order = poly_order
        assert plugin.poly_order_invalid_msg == vue_err_msg
        with pytest.raises(ValueError, match=err_msg):
            plugin.vue_add_model({})


    with pytest.raises(TraitError, match="The 'poly_order' trait of a ModelFitting instance expected an int, not the float 2.5."): # noqa
        plugin.poly_order = 2.5

    plugin.poly_order = 3
    plugin.model_comp_selected = "Polynomial1D"
    plugin.vue_add_model({})
    assert plugin.model_components[-1] == 'P3'


def test_model_ids(cubeviz_helper, spectral_cube_wcs):
    cubeviz_helper.load_data(Spectrum(flux=np.ones((3, 4, 5)) * u.nJy, wcs=spectral_cube_wcs),
                             data_label='test')
    plugin = cubeviz_helper.plugins["Model Fitting"]._obj
    plugin.dataset_selected = 'Spectrum (sum)'
    plugin.component_models = [{'id': 'valid_string_already_exists'}]
    plugin.model_comp_selected = 'Linear1D'

    with pytest.raises(
            ValueError,
            match="model component label 'valid_string_already_exists' already in use"):
        plugin.comp_label = 'valid_string_already_exists'
        plugin.vue_add_model({})

    with pytest.raises(
            ValueError,
            match="invalid model component label 'invalid-string'"):
        plugin.comp_label = 'invalid-string'
        plugin.vue_add_model({})


def test_parameter_retrieval(cubeviz_helper, spectral_cube_wcs):

    flux_unit = u.nJy
    sb_unit = flux_unit / PIX2
    wav_unit = u.Hz

    flux = np.ones((3, 4, 5))
    flux[2, 2, :] = [1, 2, 3, 4, 5]
    cubeviz_helper.load_data(Spectrum(flux=flux * flux_unit, wcs=spectral_cube_wcs),
                             data_label='test')

    plugin = cubeviz_helper.plugins["Model Fitting"]

    # since cube_fit is true, output fit should be in SB units
    # even though the spectral y axis is in 'flux' by default
    plugin.cube_fit = True

    assert cubeviz_helper.app._get_display_unit('spectral') == wav_unit
    assert cubeviz_helper.app._get_display_unit('spectral_y') == flux_unit
    assert cubeviz_helper.app._get_display_unit('sb') == sb_unit

    plugin.create_model_component("Linear1D", "L")
    # NOTE: Hardcoding n_cpu=1 to run in serial, it's slower to spool up
    # multiprocessing for the size of the cube
    plugin._obj.parallel_n_cpu = 1
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        plugin.calculate_fit()

    params = cubeviz_helper.plugins['Model Fitting'].get_model_parameters()
    slope_res = np.zeros((3, 4))
    slope_res[2, 2] = 1.0

    slope_res = slope_res * sb_unit / wav_unit
    intercept_res = np.ones((3, 4))
    intercept_res[2, 2] = 0
    intercept_res = intercept_res * sb_unit
    assert_quantity_allclose(params['model']['slope'], slope_res,
                             atol=1e-10 * sb_unit / wav_unit)
    assert_quantity_allclose(params['model']['intercept'], intercept_res,
                             atol=1e-10 * sb_unit)


@pytest.mark.parametrize(
    ('n_cpu', 'unc'), [
        (1, "zeros"),
        (1, None),
        (None, "zeros"),
        (None, None),])
def test_fitting_backend(n_cpu, unc):
    np.random.seed(42)

    x, y = build_spectrum()

    # The uncertainty array of all zero should be ignored when fitting
    if unc == "zeros":
        uncertainties = StdDevUncertainty(np.zeros(y.shape)*u.Jy)
    elif unc is None:
        uncertainties = None
    spectrum = Spectrum(flux=y*u.Jy, spectral_axis=x*u.um, uncertainty=uncertainties)

    g1f = models.Gaussian1D(0.7*u.Jy, 4.65*u.um, 0.3*u.um, name='g1')
    g2f = models.Gaussian1D(2.0*u.Jy, 5.55*u.um, 0.3*u.um, name='g2')
    g3f = models.Gaussian1D(-2.*u.Jy, 8.15*u.um, 0.2*u.um, name='g3')
    zero_level = models.Const1D(1.*u.Jy, name='const1d')

    model_list = [g1f, g2f, g3f, zero_level]
    expression = "g1 + g2 + g3 + const1d"

    # Returns the initial model
    fm, fitted_spectrum = fb.fit_model_to_spectrum(spectrum, model_list, expression,
                                                   run_fitter=False, n_cpu=n_cpu)

    parameters_expected = np.array([0.7, 4.65, 0.3, 2., 5.55, 0.3, -2.,
                                    8.15, 0.2, 1.])
    assert_allclose(fm.parameters, parameters_expected, atol=1e-5)

    # Returns the fitted model
    fm, fitted_spectrum = fb.fit_model_to_spectrum(spectrum, model_list, expression,
                                                   run_fitter=True, n_cpu=n_cpu)

    parameters_expected = np.array([1.0104705, 4.58956282, 0.19590464, 2.39892026,
                                    5.49867754, 0.10834472, -1.66902953, 8.19714439,
                                    0.09535613, 3.99125545])
    assert_allclose(fm.parameters, parameters_expected, atol=1e-5)


# For coverage of serial vs multiprocessing and unc
@pytest.mark.parametrize(
    ('n_cpu', 'unc'), [
        (1, "zeros"),
        (1, None),
        (None, "zeros"),
        (None, None)])
def test_cube_fitting_backend(cubeviz_helper, unc, n_cpu, tmp_path):
    np.random.seed(42)

    SIGMA = 0.1  # noise in data
    TOL = 0.4  # test tolerance
    IMAGE_SIZE_X = 15
    IMAGE_SIZE_Y = 14

    # Flux cube oriented as in JWST data. To build a Spectrum
    # instance with this, one need to transpose it so the spectral
    # axis direction corresponds to the last index.
    flux_cube = np.zeros((SPECTRUM_SIZE, IMAGE_SIZE_X, IMAGE_SIZE_Y))

    # Generate list of all spaxels to be fitted
    _spx = [[(x, y) for x in range(IMAGE_SIZE_X)] for y in range(IMAGE_SIZE_Y)]
    spaxels = [item for sublist in _spx for item in sublist]

    # Fill cube spaxels with spectra that differ from
    # each other only by their noise component.
    x, _ = build_spectrum()
    for spx in spaxels:
        flux_cube[:, spx[0], spx[1]] = build_spectrum(sigma=SIGMA)[1]

    # Transpose so it can be packed in a Spectrum instance.
    flux_cube = flux_cube.transpose(1, 2, 0)  # (15, 14, 200)
    cube_wcs = WCS({
        'WCSAXES': 3, 'RADESYS': 'ICRS', 'EQUINOX': 2000.0,
        'CRPIX3': 38.0, 'CRPIX2': 38.0, 'CRPIX1': 1.0,
        'CRVAL3': 205.4384, 'CRVAL2': 27.004754, 'CRVAL1': 0.0,
        'CDELT3': 0.01, 'CDELT2': 0.01, 'CDELT1': 0.05,
        'CUNIT3': 'deg', 'CUNIT2': 'deg', 'CUNIT1': 'um',
        'CTYPE3': 'RA---TAN', 'CTYPE2': 'DEC--TAN', 'CTYPE1': 'WAVE'})

    # Mask part of the spectral axis to later ensure that it gets propagated through:
    mask = np.zeros_like(flux_cube).astype(bool)
    mask[..., :SPECTRUM_SIZE // 10] = True

    # The uncertainty array of all zero should be ignored when fitting
    if unc == "zeros":
        uncertainties = StdDevUncertainty(np.zeros(flux_cube.shape)*u.Jy)
    elif unc is None:
        uncertainties = None

    spectrum = Spectrum(flux=flux_cube*u.Jy, wcs=cube_wcs,
                        uncertainty=uncertainties, mask=mask)

    # Initial model for fit.
    g1f = models.Gaussian1D(0.7*u.Jy, 4.65*u.um, 0.3*u.um, name='g1')
    g2f = models.Gaussian1D(2.0*u.Jy, 5.55*u.um, 0.3*u.um, name='g2')
    g3f = models.Gaussian1D(-2.*u.Jy, 8.15*u.um, 0.2*u.um, name='g3')
    zero_level = models.Const1D(1.*u.Jy, name='const1d')

    model_list = [g1f, g2f, g3f, zero_level]
    expression = "g1 + g2 + g3 + const1d"

    # Fit to all spaxels.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="The fit may be unsuccessful*")
        fitted_parameters, fitted_spectrum = fb.fit_model_to_spectrum(
            spectrum, model_list, expression, n_cpu=n_cpu)

    # Check that parameter results are formatted as expected.
    assert isinstance(fitted_parameters, list)
    assert len(fitted_parameters) == IMAGE_SIZE_X * IMAGE_SIZE_Y

    for m in fitted_parameters:
        if m['x'] == 3 and m['y'] == 2:
            fitted_model = m['model']

    assert isinstance(fitted_model[0].amplitude.value, np.float64)
    assert fitted_model[0].amplitude.unit == u.Jy

    assert isinstance(fitted_model[0], models.Gaussian1D)
    assert isinstance(fitted_model[0].mean.value, np.float64)
    assert fitted_model[0].mean.unit == u.um

    # Check that spectrum result is formatted as expected.
    assert isinstance(fitted_spectrum, Spectrum)
    assert len(fitted_spectrum.shape) == 3
    assert fitted_spectrum.shape == (IMAGE_SIZE_X, IMAGE_SIZE_Y, SPECTRUM_SIZE)

    # since were not loading data into app, results are just u.Jy not u.Jy / pix2
    assert fitted_spectrum.flux.unit == u.Jy
    assert not np.all(fitted_spectrum.flux.value == 0)

    # The important point here isn't to check the accuracy of the
    # fit, which was already tested elsewhere. We are mostly
    # interested here in checking the correctness of the data
    # packaging into the output products.

    assert_allclose(fitted_model[0].amplitude.value, 1.09, atol=TOL)
    assert_allclose(fitted_model[1].amplitude.value, 2.4, atol=TOL)
    assert_allclose(fitted_model[2].amplitude.value, -1.7, atol=TOL)

    assert_allclose(fitted_model[0].mean.value, 4.6, atol=TOL)
    assert_allclose(fitted_model[1].mean.value, 5.5, atol=TOL)
    assert_allclose(fitted_model[2].mean.value, 8.2, atol=TOL)

    assert_allclose(fitted_model[0].stddev.value, 0.2, atol=TOL)
    assert_allclose(fitted_model[1].stddev.value, 0.1, atol=TOL)
    assert_allclose(fitted_model[2].stddev.value, 0.1, atol=TOL)

    assert_allclose(fitted_model[3].amplitude.value, 4.0, atol=TOL)

    # Check that the fitted spectrum is masked correctly:
    assert_array_equal(fitted_spectrum.mask, mask)

    # Check I/O roundtrip.
    out_fn = tmp_path / "fitted_cube.fits"
    fitted_spectrum.write(out_fn, format="wcs1d-fits", hdu=1, flux_name="SCI", overwrite=True)
    flux_unit_str = fitted_spectrum.flux.unit.to_string(format="fits")
    coo_expected = fitted_spectrum.wcs.pixel_to_world(1, 0, 2)
    with fits.open(out_fn) as pf:
        assert len(pf) == 3
        assert pf[0].name == "PRIMARY"
        assert pf[1].name == "SCI"
        assert pf[1].header["BUNIT"] == flux_unit_str
        assert_allclose(pf[1].data, fitted_spectrum.flux.value)
        assert pf[2].name == "MASK"
        assert_array_equal(pf[2].data, mask)
        w = WCS(pf[1].header)
        coo = w.pixel_to_world(1, 0, 2)
        assert_allclose(coo[0].value, coo_expected[0].value)  # SpectralCoord
        assert_allclose([coo[1].ra.deg, coo[1].dec.deg],
                        [coo_expected[1].ra.deg, coo_expected[1].dec.deg])

    # Check Cubeviz roundtrip. This should automatically go through wcs1d-fits reader.
    cubeviz_helper.load_data(out_fn)
    assert len(cubeviz_helper.app.data_collection) == 3
    data_sci = cubeviz_helper.app.data_collection["fitted_cube"]
    flux_sci = data_sci.get_component("flux")
    assert_allclose(flux_sci.data, fitted_spectrum.flux.value)
    # now that the flux cube was loaded into cubeviz, there will be a factor
    # of pix2 applied to the flux unit
    assert flux_sci.units == f'{flux_unit_str} / pix2'
    coo = data_sci.coords.pixel_to_world(1, 0, 2)
    assert_allclose(coo[0].value, coo_expected[0].value)  # SpectralCoord
    assert_allclose([coo[1].ra.deg, coo[1].dec.deg],
                    [coo_expected[1].ra.deg, coo_expected[1].dec.deg])
    data_mask = cubeviz_helper.app.data_collection["fitted_cube [MASK]"]
    flux_mask = data_mask.get_component("flux")
    assert_array_equal(flux_mask.data, mask)


def test_results_table(specviz_helper, spectrum1d):
    data_label = 'test'
    specviz_helper.load_data(spectrum1d, data_label=data_label)

    mf = specviz_helper.plugins['Model Fitting']
    mf.create_model_component('Linear1D')

    uc = specviz_helper.plugins['Unit Conversion']

    mf.add_results.label = 'linear model'
    mf._obj.parallel_n_cpu = 1
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit(add_data=True)
    mf_table = mf.export_table()
    assert len(mf_table) == 1
    assert mf_table['equation'][-1] == 'L'
    assert mf_table.colnames == ['model', 'data_label', 'spectral_subset', 'equation',
                                 'L:slope_0', 'L:slope_0:unit',
                                 'L:slope_0:fixed', 'L:slope_0:std',
                                 'L:intercept_0', 'L:intercept_0:unit',
                                 'L:intercept_0:fixed', 'L:intercept_0:std']

    # verify units in table match the current display unit
    assert mf_table['L:intercept_0:unit'][-1] == uc.flux_unit

    mf.create_model_component('Gaussian1D')
    mf.add_results.label = 'composite model'
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        warnings.filterwarnings('ignore', message='The fit may be unsuccessful*')
        mf.calculate_fit(add_data=True)
    mf_table = mf.export_table()
    assert len(mf_table) == 2
    assert mf_table['equation'][-1] == 'L+G'
    assert mf_table.colnames == ['model', 'data_label', 'spectral_subset', 'equation',
                                 'L:slope_0', 'L:slope_0:unit',
                                 'L:slope_0:fixed', 'L:slope_0:std',
                                 'L:intercept_0', 'L:intercept_0:unit',
                                 'L:intercept_0:fixed', 'L:intercept_0:std',
                                 'G:amplitude_1', 'G:amplitude_1:unit',
                                 'G:amplitude_1:fixed', 'G:amplitude_1:std',
                                 'G:mean_1', 'G:mean_1:unit',
                                 'G:mean_1:fixed', 'G:mean_1:std',
                                 'G:stddev_1', 'G:stddev_1:unit',
                                 'G:stddev_1:fixed', 'G:stddev_1:std']

    mf.remove_model_component('G')
    assert len(mf_table) == 2

    # verify Spectrum model fitting plugin and table can handle spectral density conversions
    uc.flux_unit = 'erg / (Angstrom s cm2)'
    mf.reestimate_model_parameters()

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit(add_data=True)

    assert mf_table['L:intercept_0:unit'][-1] == uc.flux_unit


def test_equation_validation(specviz_helper, spectrum1d):
    data_label = 'test'
    specviz_helper.load_data(spectrum1d, data_label=data_label)

    mf = specviz_helper.plugins['Model Fitting']
    mf.create_model_component('Const1D')
    mf.create_model_component('Linear1D')

    assert mf.equation == 'C+L'
    assert mf._obj.model_equation_invalid_msg == ''

    mf.equation = 'L+'
    assert mf._obj.model_equation_invalid_msg == 'incomplete equation.'

    mf.equation = 'L+C'
    assert mf._obj.model_equation_invalid_msg == ''

    mf.equation = 'L+CC'
    assert mf._obj.model_equation_invalid_msg == 'CC is not an existing model component.'

    mf.equation = 'L+CC+DD'
    assert mf._obj.model_equation_invalid_msg == 'CC, DD are not existing model components.'

    mf.equation = ''
    assert mf._obj.model_equation_invalid_msg == 'model equation is required.'


def test_incompatible_units(specviz_helper, spectrum1d):
    data_label = 'test'
    specviz_helper.load_data(spectrum1d, data_label=data_label)

    mf = specviz_helper.plugins['Model Fitting']
    mf.create_model_component('Linear1D')

    mf.add_results.label = 'model native units'
    mf._obj.parallel_n_cpu = 1
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit(add_data=True)

    uc = specviz_helper.plugins['Unit Conversion']
    assert uc.spectral_unit.selected == "Angstrom"
    uc.spectral_unit = u.Hz

    assert 'L is currently disabled' in mf._obj.model_equation_invalid_msg
    mf.add_results.label = 'frequency units'
    with pytest.raises(ValueError, match="model equation is invalid.*"):
        mf.calculate_fit(add_data=True)

    mf.reestimate_model_parameters()
    assert mf._obj.model_equation_invalid_msg == ''
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit(add_data=True)


def test_cube_fit_with_nans(cubeviz_helper):
    flux = np.ones((7, 8, 9)) * u.nJy
    flux[:, :, 0] = np.nan
    spec = Spectrum(flux=flux, spectral_axis_index=2)
    cubeviz_helper.load_data(spec, data_label="test")

    mf = cubeviz_helper.plugins["Model Fitting"]
    mf.cube_fit = True
    mf.create_model_component("Const1D")
    mf._obj.parallel_n_cpu = 1

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit()

    result = cubeviz_helper.app.data_collection['model']
    assert np.all(result.get_component("flux").data == 1)

    # Switch back to non-cube fit, check that units are marked incompatible
    mf.cube_fit = False
    assert mf._obj.component_models[0]['compat_display_units'] is False


def test_cube_fit_with_subset_and_nans(cubeviz_helper):
    # Also test with existing mask
    flux = np.ones((7, 8, 9)) * u.nJy
    flux[:, :, 0] = np.nan
    spec = Spectrum(flux=flux, spectral_axis_index=2)
    spec.flux[5, 5, 7] = 10 * u.nJy
    cubeviz_helper.load_data(spec, data_label="test")

    sv = cubeviz_helper.app.get_viewer('spectrum-viewer')
    sv.apply_roi(XRangeROI(0, 5))

    mf = cubeviz_helper.plugins["Model Fitting"]
    mf.cube_fit = True
    mf.spectral_subset = 'Subset 1'
    mf.create_model_component("Const1D")
    mf._obj.parallel_n_cpu = 1

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit()
    result = cubeviz_helper.app.data_collection['model']
    assert np.all(result.get_component("flux").data == 1)


def test_fit_with_count_units(cubeviz_helper):
    flux = np.random.random((7, 8, 9)) * u.count
    spectral_axis = np.linspace(4000, 5000, flux.shape[-1]) * u.AA

    spec = Spectrum(flux=flux, spectral_axis=spectral_axis, spectral_axis_index=2)
    cubeviz_helper.load_data(spec, data_label="test")

    mf = cubeviz_helper.plugins["Model Fitting"]
    mf.cube_fit = True
    mf.create_model_component("Const1D")
    mf._obj.parallel_n_cpu = 1

    # ensures specutils.Spectrum.with_flux_unit has access to Jdaviz custom equivalencies for
    # PIX^2 unit
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit()

    assert mf._obj.component_models[0]['parameters'][0]['unit'] == 'ct / pix2'

    model_flux = cubeviz_helper.app.data_collection[-1].get_component('flux')
    assert model_flux.units == 'ct / pix2'


@pytest.mark.parametrize("solid_angle_unit", [u.sr, PIX2])
def test_cube_fit_after_unit_change(cubeviz_helper, solid_angle_unit):
    cube = _create_spectrum1d_cube_with_fluxunit(fluxunit=u.Jy / solid_angle_unit, shape=(10, 4, 5),
                                                 with_uncerts=True)
    cubeviz_helper.load_data(cube, data_label="test")
    solid_angle_string = str(solid_angle_unit)

    uc = cubeviz_helper.plugins['Unit Conversion']
    mf = cubeviz_helper.plugins['Model Fitting']
    uc.flux_unit = "MJy"
    mf.cube_fit = True

    mf.create_model_component("Const1D")
    # ensure parameters amplitude unit matches display unit
    assert mf.get_model_component('C').get('parameters').get('amplitude').get('unit') == f'MJy / {solid_angle_string}'  # noqa
    # Check that the parameter is using the current units when initialized
    assert mf._obj.component_models[0]['parameters'][0]['unit'] == f'MJy / {solid_angle_string}'

    mf._obj.parallel_n_cpu = 1
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit()

    # It was easier to transpose this for new data shape than rewrite it
    expected_result_slice = np.array([[9.00e-05, 9.50e-05, 1.00e-04, 1.05e-04],
                                      [9.10e-05, 9.60e-05, 1.01e-04, 1.06e-04],
                                      [9.20e-05, 9.70e-05, 1.02e-04, 1.07e-04],
                                      [9.30e-05, 9.80e-05, 1.03e-04, 1.08e-04],
                                      [9.40e-05, 9.90e-05, 1.04e-04, 1.09e-04]]).T

    model_flux = cubeviz_helper.app.data_collection[-1].get_component('flux')
    assert model_flux.units == f'MJy / {solid_angle_string}'
    assert np.allclose(model_flux.data[1, :, :], expected_result_slice)

    # Switch back to Jy, see that the component didn't change but the output does
    uc.flux_unit = 'Jy'
    assert mf._obj.component_models[0]['parameters'][0]['unit'] == f'Jy / {solid_angle_string}'
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit()

    model_flux = cubeviz_helper.app.data_collection[-1].get_component('flux')
    assert model_flux.units == f'Jy / {solid_angle_string}'
    assert np.allclose(model_flux.data[1, :, :], expected_result_slice * 1e6)

    # ensure conversions that require the spectral axis/translations are handled by the plugin
    uc.spectral_y_type = 'Surface Brightness'
    uc.flux_unit = 'erg / (Angstrom s cm2)'

    mf.reestimate_model_parameters()

    if solid_angle_string == 'sr':
        expected_unit_string = f'erg / (Angstrom s {solid_angle_string} cm2)'
        assert mf.get_model_component('C').get('parameters').get('amplitude').get('unit') == expected_unit_string  # noqa
    else:
        expected_unit_string = f'erg / (Angstrom s cm2 {solid_angle_string})'
        assert mf.get_model_component('C').get('parameters').get('amplitude').get('unit') == expected_unit_string  # noqa

    assert mf._obj.component_models[0]['parameters'][0]['unit'] == expected_unit_string

    # running this ensures specutils.Spectrum.with_flux_unit has Jdaviz custom equivalencies
    # for spectral axis conversions and scale factor translations
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit()

    model_flux = cubeviz_helper.app.data_collection[-1].get_component('flux')
    assert model_flux.units == expected_unit_string


@pytest.mark.parametrize('op, allowed',[('+', True),('-', True),('*', False),('/', False),('**', False),],)  # noqa
def test_valid_component_operators(deconfigged_helper, op, allowed):
    flux = np.ones(9) * u.Jy
    flux[7] = 10 * u.Jy
    wavelength = np.arange(9) * u.um
    spec = Spectrum(flux=flux, spectral_axis=wavelength)
    deconfigged_helper.load(
        spec, data_label='1D Spectrum', format='1D Spectrum'
    )

    mf = deconfigged_helper.plugins['Model Fitting']
    mf.create_model_component('Const1D')
    mf.create_model_component('Const1D')

    mf.equation.value = f'C{op}C_1'

    if allowed:
        assert mf._obj.model_equation_invalid_msg == ''
    else:
        assert 'unsupported operator' in mf._obj.model_equation_invalid_msg
        assert op in mf._obj.model_equation_invalid_msg


def test_deconf_mf_with_subset(deconfigged_helper):
    flux = np.ones(9) * u.Jy
    flux[7] = 10 * u.Jy
    wavelength = np.arange(9) * u.um
    spec = Spectrum(flux=flux, spectral_axis=wavelength)
    deconfigged_helper.load(spec, data_label="1D Spectrum", format='1D Spectrum')

    subset = deconfigged_helper.plugins['Subset Tools']
    subset.import_region(SpectralRegion(6 * u.Unit('um'), 7 * u.Unit('um')))

    mf = deconfigged_helper.plugins['Model Fitting']
    # ensure deconfigged app can access subsets in plugin
    mf.spectral_subset.selected = 'Subset 1'
    mf.add_results.label = 'linear model'
    mf._obj.parallel_n_cpu = 1
    mf.create_model_component()

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit()


@pytest.mark.parametrize('fitter', ['TRFLSQFitter', 'DogBoxLSQFitter', 'LMLSQFitter',
                                    'LevMarLSQFitter', 'LinearLSQFitter', 'SLSQPLSQFitter',
                                    'SimplexLSQFitter'])
def test_different_fitters(specviz_helper, spectrum1d, fitter):
    data_label = 'test'
    specviz_helper.load_data(spectrum1d, data_label=data_label)

    mf = specviz_helper.plugins['Model Fitting']._obj
    if fitter == 'SimplexLSQFitter':
        mf.create_model_component('Const1D')
    else:
        mf.create_model_component('Linear1D')

    mf.fitter.selected = fitter
    # change maxiter to 50
    if fitter != 'LinearLSQFitter':
        mf.fitter_parameters['parameters'][0]['value'] = 50

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit(add_data=True)

    result = specviz_helper.app.data_collection['model']
    expected_result = [6000., 6222.22222222, 6444.44444444, 6666.66666667, 6888.88888889,
                       7111.11111111, 7333.33333333, 7555.55555556, 7777.77777778, 8000.] * u.AA
    assert_allclose(result.get_object().spectral_axis, expected_result)


def test_specviz2d_linking(specviz2d_helper):
    header = {
              'WCSAXES': 2,
              'CRPIX1': 0.0, 'CRPIX2': 8.5,
              'CDELT1': 1E-06, 'CDELT2': 7.5E-05,
              'CUNIT1': 'm', 'CUNIT2': 'deg',
              'CTYPE1': 'WAVE', 'CTYPE2': 'OFFSET',
              'CRVAL1': 0.0, 'CRVAL2': 5.0,
              'RADESYS': 'ICRS', 'SPECSYS': 'BARYCENT'}
    wcs = WCS(header)

    x_values = np.linspace(0, 10, 128)
    y_values = np.linspace(0, 5, 256)

    # Create a continuous 2D
    data = np.sin(x_values[:, np.newaxis]) * np.cos(y_values) * u.one
    spectrum_data = Spectrum(data, wcs=wcs, meta=header)
    specviz2d_helper.load_data(spectrum_2d=spectrum_data)

    viewer_1d = specviz2d_helper.app.get_viewer(
        specviz2d_helper._default_spectrum_viewer_reference_name)

    mf = specviz2d_helper.plugins['Model Fitting']
    mf.create_model_component('Const1D')
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit()

    # testing to ensure a bug where the fit values create a vertical line along the
    # x-axis (x=0). This test verifies that the x-range is of the fit is within the
    # limits of the spectrum data used to produce the fit.
    fit_marks_x = viewer_1d.native_marks[-1].x
    fit_x_min = np.min(fit_marks_x)
    fit_x_max = np.max(fit_marks_x)

    data_marks_x = viewer_1d.native_marks[0].x
    data_x_min = np.min(data_marks_x)
    data_x_max = np.max(data_marks_x)

    assert data_x_min <= fit_x_min <= fit_x_max <= data_x_max, (
        f"Fit range [{fit_x_min},{fit_x_max}] isn't within data range [{data_x_min},{data_x_max}]"
    )


def test_model_equation_with_different_flux_units(specviz_helper):
    wavelength = np.linspace(6000, 8000, 100) * u.AA
    # Define individual model components
    # Linear model: y = 2.0 * x (slope=2.0)
    linear_component = 2.0 * wavelength.value
    # Constant model: y = 500 (constant=500)
    constant_component = 500.0
    # Polynomial model: y = 0.0001 * x^2 - 0.5 * x + 100 (c0=100, c1=-0.5, c2=0.0001)
    poly_component = 0.0001 * wavelength.value**2 - 0.5 * wavelength.value + 100
    # Combine all components
    flux = (linear_component + constant_component + poly_component) * u.Jy
    # Add uncertainty
    uncertainty = StdDevUncertainty(np.ones_like(flux.value) * 0.1 * u.Jy)
    # Create spectrum
    spec = Spectrum(spectral_axis=wavelength, flux=flux, uncertainty=uncertainty)
    data_label = 'test'
    specviz_helper.load_data(spec, data_label=data_label)

    mf = specviz_helper.plugins['Model Fitting']
    uc = specviz_helper.plugins['Unit Conversion']

    mf._obj.fitter.selected = 'LevMarLSQFitter'

    # Create first model component with flux unit MJy
    uc.flux_unit = 'MJy'
    mf.model_component = 'Linear1D'
    mf.create_model_component()

    # Make sure unit conversion is correct for model component params
    assert_allclose(mf._obj.component_models[0]['parameters'][0]['value'], 0.016034, rtol=1e-5)
    assert_allclose(mf._obj.component_models[0]['parameters'][1]['value'], 0.0029, rtol=1e-5)

    # Set initial values to be exact answer to test if reestimate works
    # mf._obj.component_models[0]['parameters'][0]['value'] = 0.002
    # mf._obj.component_models[0]['parameters'][1]['value'] = 0.001

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit(add_data=True)

    # Make sure the slope values and units are updating correctly
    assert mf._obj.component_models[0]['parameters'][0]['unit'] == 'MJy / Angstrom'
    assert_allclose(mf._obj.component_models[0]['parameters'][0]['value'], 2.90e-06)
    assert_allclose(mf._obj.component_models[0]['parameters'][0]['std'], 1.714816e-11, rtol=1e-5)
    # Make sure the intercept values and units are updating correctly
    assert mf._obj.component_models[0]['parameters'][1]['unit'] == 'MJy'
    assert_allclose(mf._obj.component_models[0]['parameters'][1]['value'], -0.004266, rtol=1e-5)
    assert_allclose(mf._obj.component_models[0]['parameters'][1]['std'], 1.204529e-07, rtol=1e-5)

    # Create second model component with flux unit Jy
    uc.flux_unit = 'Jy'
    mf.model_component = 'Const1D'
    mf.create_model_component()

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit(add_data=True)

    # Testing the individual parameter values and units in the component models
    # was not working because after the calculate_fit call, the Linear1D models
    # would add together to equal the correct slope and intercept. Checking the individual
    # model component parameters after this call would not work because the values could be
    # different depending on the environment. So instead we just check the units here and
    # make sure the final model matches the input spectrum.

    # Make sure the slope units are updating correctly
    assert mf._obj.component_models[0]['parameters'][0]['unit'] == 'Jy / Angstrom'

    # Create third model component with flux unit W / (Hz m2)
    uc.flux_unit = 'W / (Hz m2)'
    mf.model_component = 'Polynomial1D'
    mf.create_model_component()

    # Make sure unit conversion is correct for model component params

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message='Model is linear in parameters*')
        mf.calculate_fit(add_data=True)

    # Make sure the slope units are updating correctly
    assert mf._obj.component_models[0]['parameters'][0]['unit'] == 'W / (Angstrom Hz m2)'

    model = specviz_helper.app.data_collection['model'].get_object(Spectrum)
    assert model.flux.unit == 'W / (Hz m2)'
    assert_allclose(model.flux, spec.flux.to('W / (Hz m2)'), rtol=1e-2)


def test_get_fitter_parameter_all_fitters(specviz_helper, spectrum1d):
    """Test the get_fitter_parameter method for all non-spline fitters."""
    fitter_configs = [
        ('TRFLSQFitter', {'maxiter': 100, 'filter_non_finite': True,
                          'calc_uncertainties': True}),
        ('DogBoxLSQFitter', {'maxiter': 100, 'filter_non_finite': True,
                             'calc_uncertainties': True}),
        ('LMLSQFitter', {'maxiter': 100, 'filter_non_finite': True,
                         'calc_uncertainties': True}),
        ('LevMarLSQFitter', {'maxiter': 100, 'filter_non_finite': True,
                             'calc_uncertainties': True}),
        ('LinearLSQFitter', {'calc_uncertainties': True}),
        ('SLSQPLSQFitter', {'maxiter': 100}),
        ('SimplexLSQFitter', {'maxiter': 100}),
    ]

    specviz_helper.load_data(spectrum1d)
    plugin = specviz_helper.plugins["Model Fitting"]

    for fitter_name, expected_params in fitter_configs:
        # Set the fitter
        plugin.fitter = fitter_name

        # Test getting each expected parameter and verify default value
        for param_name, expected_value in expected_params.items():
            actual_value = plugin.get_fitter_parameter(param_name)
            assert actual_value == expected_value, (
                f"Fitter {fitter_name}: expected {param_name}={expected_value}, "
                f"got {actual_value}"
            )

        # Test getting non-existent parameter returns None
        non_existent = plugin.get_fitter_parameter('non_existent_param')
        assert non_existent is None

        # Test that parameters not in this fitter return None
        if 'maxiter' not in expected_params:
            assert plugin.get_fitter_parameter('maxiter') is None
        if 'filter_non_finite' not in expected_params:
            assert plugin.get_fitter_parameter('filter_non_finite') is None


def test_set_fitter_parameter_all_fitters(specviz_helper, spectrum1d):
    """Test the set_fitter_parameter method for all non-spline fitters."""
    fitter_configs = [
        ('TRFLSQFitter', {'maxiter': 50, 'filter_non_finite': False,
                          'calc_uncertainties': False}),
        ('DogBoxLSQFitter', {'maxiter': 75, 'filter_non_finite': False,
                             'calc_uncertainties': False}),
        ('LMLSQFitter', {'maxiter': 125, 'filter_non_finite': False,
                         'calc_uncertainties': False}),
        ('LevMarLSQFitter', {'maxiter': 150, 'filter_non_finite': False,
                             'calc_uncertainties': False}),
        ('LinearLSQFitter', {'calc_uncertainties': False}),
        ('SLSQPLSQFitter', {'maxiter': 200}),
        ('SimplexLSQFitter', {'maxiter': 250}),
    ]

    specviz_helper.load_data(spectrum1d)
    plugin = specviz_helper.plugins["Model Fitting"]

    for fitter_name, test_values in fitter_configs:
        # Set the fitter
        plugin.fitter = fitter_name

        # Test setting each parameter and verify it was set correctly
        for param_name, test_value in test_values.items():
            plugin.set_fitter_parameter(param_name, test_value)
            actual_value = plugin.get_fitter_parameter(param_name)
            assert actual_value == test_value, (
                f"Fitter {fitter_name}: failed to set {param_name}={test_value}, "
                f"got {actual_value}"
            )

        # Test that setting non-existent parameter doesn't raise error
        plugin.set_fitter_parameter('non_existent_param', 999)
        assert plugin.get_fitter_parameter('non_existent_param') is None


def test_spline_fitter_get_set_parameters(specviz_helper, spectrum1d):
    """Test get/set fitter parameters for SplineSmoothingFitter."""
    specviz_helper.load_data(spectrum1d)
    plugin = specviz_helper.plugins["Model Fitting"]

    # Create Spline1D model to enable SplineSmoothingFitter
    plugin.create_model_component('Spline1D')
    plugin.fitter = 'SplineSmoothingFitter'

    # Test getting default parameters
    assert plugin.get_fitter_parameter('maxiter') == 100
    assert plugin.get_fitter_parameter('smoothing_factor') == 20.0
    assert plugin.get_fitter_parameter('degree') == 3.0

    # Test setting parameters
    plugin.set_fitter_parameter('maxiter', 50)
    assert plugin.get_fitter_parameter('maxiter') == 50

    plugin.set_fitter_parameter('smoothing_factor', 0.5)
    assert plugin.get_fitter_parameter('smoothing_factor') == 0.5

    plugin.set_fitter_parameter('degree', 4.0)
    assert plugin.get_fitter_parameter('degree') == 4.0

    # Test non-existent parameter
    assert plugin.get_fitter_parameter('non_existent') is None
    plugin.set_fitter_parameter('non_existent', 123)
    assert plugin.get_fitter_parameter('non_existent') is None


def test_fitter_parameter_persistence(specviz_helper, spectrum1d):
    """Test that fitter parameters persist when switching between fitters."""
    specviz_helper.load_data(spectrum1d)
    plugin = specviz_helper.plugins["Model Fitting"]

    # Set custom values for multiple fitters
    fitter_configs = {
        'LevMarLSQFitter': {'maxiter': 75},
        'SimplexLSQFitter': {'maxiter': 150},
        'TRFLSQFitter': {'maxiter': 225, 'filter_non_finite': False},
        'LinearLSQFitter': {'calc_uncertainties': False},
    }

    # Set all configurations
    for fitter_name, params in fitter_configs.items():
        plugin.fitter = fitter_name
        for param_name, param_value in params.items():
            plugin.set_fitter_parameter(param_name, param_value)

    # Verify all configurations persisted
    for fitter_name, params in fitter_configs.items():
        plugin.fitter = fitter_name
        for param_name, expected_value in params.items():
            actual_value = plugin.get_fitter_parameter(param_name)
            assert actual_value == expected_value, (
                f"Fitter {fitter_name}: parameter {param_name} did not persist, "
                f"expected {expected_value}, got {actual_value}"
            )


def test_spline(specviz_helper, spectrum1d):
    data_label = 'test'
    specviz_helper.load_data(spectrum1d, data_label=data_label)
    mf = specviz_helper.plugins['Model Fitting']._obj
    mf.create_model_component('Spline1D')

    mf.fitter.selected = 'SplineSmoothingFitter'

    # check that smoothing_factor parameter exists and has a value (initialized as 0
    # by unitl the fitter component is selected)
    assert mf.fitter_parameters['parameters'][1]['name'] == 'smoothing_factor'
    assert mf.fitter_parameters['parameters'][1]['value'] is not None

    deg_param = next(p for p in mf.fitter_parameters['parameters']
                     if p['name'] == 'degree')

    # degree must be between 1 and 5
    deg_param['value'] = 0
    with pytest.raises(ValueError, match=r"k should be 1 <= k <= 5"):
        mf.calculate_fit(add_data=False)
    deg_param['value'] = 6
    with pytest.raises(ValueError, match=r"k should be 1 <= k <= 5"):
        mf.calculate_fit(add_data=False)

    deg_param['value'] = 3

    mf.calculate_fit(add_data=True)

    # ensure that Spline1D is not combined with any other model components
    mf.create_model_component('Const1D')
    assert mf.model_equation_invalid_msg == (
                    "Spline1D cannot be combined with other model components."
                )
    mf.remove_model_component('C')

    # ensure that only SplineSmoothingFitter fitter component can be used
    # with the Spline1D model parameter
    assert mf.fitter.choices == ['SplineSmoothingFitter']

    mf.remove_model_component('S')
    mf.create_model_component('Const1D')

    # make sure fitter components update when Spline1D model component is removed
    assert mf.fitter.choices == ['TRFLSQFitter',
                                 'DogBoxLSQFitter',
                                 'LMLSQFitter',
                                 'LevMarLSQFitter',
                                 'LinearLSQFitter',
                                 'SLSQPLSQFitter',
                                 'SimplexLSQFitter',
                                 'SplineSmoothingFitter']



from astropy.io import fits
from astropy.nddata import NDData
import astropy.units as u
import pytest
import numpy as np
from specutils import SpectralRegion
from glue.core.roi import (
    RectangularROI, CircularROI, EllipticalROI,
    CircularAnnulusROI, XRangeROI
)
from glue.core.edit_subset_mode import ReplaceMode


def test_data_menu_toggles(specviz_helper, spectrum1d):
    # NOTE: this test is adopted from core.tests.test_data_menu.test_data_menu_toggles
    # which should be removed once the old data menu is removed from jdaviz

    # load 2 data entries
    specviz_helper.load_data(spectrum1d, data_label="test")
    new_spec = specviz_helper.get_spectra(apply_slider_redshift=True)["test"]*0.9
    specviz_helper.load_data(new_spec, data_label="test2")

    # check that both are enabled in the data menu
    sv = specviz_helper.viewers['spectrum-viewer']
    dm = sv._obj.glue_viewer.data_menu
    assert len(dm._obj.layer_items) == 2
    assert len(dm._obj.visible_layers) == 2

    # disable (hide layer) for second entry
    dm.set_layer_visibility('test2', False)
    assert len(dm._obj.layer_items) == 2
    assert len(dm._obj.visible_layers) == 1

    # add a subset and make sure it appears for the first data entry but not the second
    specviz_helper.plugins['Subset Tools'].import_region(
        SpectralRegion(6000 * spectrum1d.spectral_axis.unit, 6500 * spectrum1d.spectral_axis.unit))

    assert len(dm._obj.layer_items) == 3
    assert len(dm._obj.visible_layers) == 2
    assert len(sv._obj.glue_viewer.layers) == 4
    assert sv._obj.glue_viewer.layers[2].visible is True   # subset corresponding to first (visible)
    assert sv._obj.glue_viewer.layers[3].visible is False  # subset corresponding to second (hidden)

    # enable data layer from menu and subset should also become visible
    dm.toggle_layer_visibility('test2')
    assert np.all([layer.visible for layer in sv._obj.glue_viewer.layers])


def test_data_menu_selection(specviz_helper, spectrum1d):
    # load 2 data entries
    specviz_helper.load_data(spectrum1d, data_label="test")
    new_spec = specviz_helper.get_spectra(apply_slider_redshift=True)["test"]*0.9
    specviz_helper.load_data(new_spec, data_label="test2")

    sv = specviz_helper.viewers['spectrum-viewer']
    dm = sv._obj.glue_viewer.data_menu

    # no selection by default
    assert len(dm._obj.dm_layer_selected) == 0
    assert len(dm.layer.selected) == 0

    # test layer -> UI sync
    dm.layer.selected = dm.layer.choices[0]
    assert len(dm._obj.dm_layer_selected) == len(dm.layer.selected)

    # test UI -> layer sync
    dm._obj.dm_layer_selected = [0, 1]
    assert len(dm._obj.dm_layer_selected) == len(dm.layer.selected)

    # test that sync remains during layer deletion
    dm._obj.dm_layer_selected = [1]
    assert dm.layer.selected == ['test']
    specviz_helper.app.remove_data_from_viewer('spectrum-viewer', "test2")
    specviz_helper.app.data_item_remove("test2")
    assert len(dm._obj.layer_items) == 1
    assert dm._obj.dm_layer_selected == [0]
    assert dm.layer.selected == ['test']


def test_data_menu_add_remove_data(imviz_helper):
    for i in range(3):
        imviz_helper.load_data(np.zeros((2, 2)) + i, data_label=f'image_{i}', show_in_viewer=False)

    dm = imviz_helper.viewers['imviz-0']._obj.glue_viewer.data_menu
    assert len(dm._obj.layer_items) == 0
    assert len(dm.layer.choices) == 0
    assert len(dm._obj.dataset.choices) == 3

    dm.add_data('image_0')
    assert dm.layer.choices == ['image_0']
    assert len(dm._obj.dataset.choices) == 2

    with pytest.raises(ValueError,
                       match="Data labels \\['dne1', 'dne2'\\] not able to be loaded into 'imviz-0'.  Must be one of: \\['image_1', 'image_2'\\]"):  # noqa
        dm.add_data('dne1', 'dne2')

    dm.add_data('image_1', 'image_2')
    assert len(dm.layer.choices) == 3
    assert len(dm._obj.dataset.choices) == 0

    dm.layer.selected = ['image_0']
    dm.remove_from_viewer()
    assert len(dm.layer.choices) == 2
    assert len(dm._obj.dataset.choices) == 1

    dm.layer.selected = ['image_1']
    dm.remove_from_app()
    assert len(dm.layer.choices) == 1
    assert len(dm._obj.dataset.choices) == 1


def test_remove_all_data_cubeviz(cubeviz_helper, image_cube_hdu_obj_microns):
    cubeviz_helper.load_data(image_cube_hdu_obj_microns, data_label="test")
    dm1 = cubeviz_helper.viewers['flux-viewer']._obj.glue_viewer.data_menu
    dm2 = cubeviz_helper.viewers['uncert-viewer']._obj.glue_viewer.data_menu
    dm3 = cubeviz_helper.viewers['spectrum-viewer']._obj.glue_viewer.data_menu

    sp = cubeviz_helper.plugins['Subset Tools']
    sp.import_region(SpectralRegion(4.8 * u.um,
                                    5 * u.um),
                     combination_mode='new')

    # Remove all data from all viewers to make sure plugins respond appropriately
    # This would previously raise a traceback from the data menu and/or aperture photometry
    dm1.layer.selected = ['test[FLUX]']
    dm1.remove_from_viewer()
    dm2.layer.selected = ['test[ERR]']
    dm2.remove_from_viewer()
    dm3.layer.selected = ['Spectrum (sum)', 'Subset 1']
    dm3.remove_from_viewer()

    for plg in ('Moment Maps', 'Line Analysis', 'Line Lists'):
        assert "unavailable" in cubeviz_helper.plugins[plg]._obj.disabled_msg


def test_data_menu_create_subset(imviz_helper):
    imviz_helper.load_data(np.zeros((2, 2)), data_label='image', show_in_viewer=True)

    dm = imviz_helper.viewers['imviz-0']._obj.glue_viewer.data_menu
    assert imviz_helper.app.session.edit_subset_mode.edit_subset == []

    dm.create_subset('circle')
    assert imviz_helper.app.session.edit_subset_mode.edit_subset == []
    assert imviz_helper.viewers['imviz-0']._obj.glue_viewer.toolbar.active_tool_id == 'bqplot:truecircle'  # noqa


def test_data_menu_remove_subset(specviz_helper, spectrum1d):
    # load 2 data entries
    specviz_helper.load_data(spectrum1d, data_label="test")
    new_spec = specviz_helper.get_spectra(apply_slider_redshift=True)["test"]*0.9
    specviz_helper.load_data(new_spec, data_label="test2")

    dm = specviz_helper.viewers['spectrum-viewer']._obj.glue_viewer.data_menu
    sp = specviz_helper.plugins['Subset Tools']

    sp.import_region(SpectralRegion(6000 * spectrum1d.spectral_axis.unit,
                                    6100 * spectrum1d.spectral_axis.unit),
                     combination_mode='new')
    sp.import_region(SpectralRegion(6000 * spectrum1d.spectral_axis.unit,
                                    6100 * spectrum1d.spectral_axis.unit),
                     combination_mode='new')

    assert dm.layer.choices == ['Subset 2', 'Subset 1', 'test2', 'test']
    dm.layer.selected = ['Subset 1']
    dm.remove_from_viewer()

    # subset visibility is set to false, but still appears in menu (unlike removing data)
    assert dm.layer.choices == ['Subset 2', 'Subset 1', 'test2', 'test']
    assert dm._obj.layer_items[1]['label'] == 'Subset 1'
    # TODO: sometimes appearing as mixed right now, known bug
    assert dm._obj.layer_items[1]['visible'] is not True

    # selection should not have changed by removing subset from viewer
    assert dm.layer.selected == ['Subset 1']
    dm.remove_from_app()
    # TODO: not quite sure why this isn't passing, seems to
    # work on local tests, so may just be async?
    # assert dm.layer.choices == ['test', 'test2', 'Subset 2']


def test_data_menu_dq_layers(imviz_helper):
    hdu_data = fits.ImageHDU(np.zeros(shape=(2, 2)))
    hdu_data.name = 'SCI'
    hdu_dq = fits.ImageHDU(np.zeros(shape=(2, 2), dtype=np.int32))
    hdu_dq.name = 'DQ'
    data = fits.HDUList([fits.PrimaryHDU(), hdu_data, hdu_dq])

    imviz_helper.load_data(data, data_label="image", ext=('SCI', 'DQ'), show_in_viewer=True)

    dm = imviz_helper.viewers['imviz-0']._obj.glue_viewer.data_menu
    assert dm.layer.choices == ['image[SCI,1]', 'image[DQ,1]']
    assert len(dm._obj.visible_layers) == 2

    # turning off image (parent) data-layer should also turn off DQ
    dm.set_layer_visibility('image[SCI,1]', False)
    assert len(dm._obj.visible_layers) == 0

    # turning on image (parent) should leave DQ off
    dm.set_layer_visibility('image[SCI,1]', True)
    assert len(dm._obj.visible_layers) == 1

    # turning on DQ (child, when parent off) should show parent
    dm.set_layer_visibility('image[SCI,1]', False)
    dm.set_layer_visibility('image[DQ,1]', True)
    assert len(dm._obj.visible_layers) == 2


@pytest.mark.xfail(reason="known issue")
def test_data_menu_subset_appearance(specviz_helper, spectrum1d):
    # NOTE: this test is similar to above - the subset is appearing in time IF there
    # are two data entries, but not in this case with just one
    specviz_helper.load_data(spectrum1d, data_label="test")

    dm = specviz_helper.viewers['spectrum-viewer']._obj.glue_viewer.data_menu
    sp = specviz_helper.plugins['Subset Tools']

    sp.import_region(SpectralRegion(6000 * spectrum1d.spectral_axis.unit,
                                    6100 * spectrum1d.spectral_axis.unit))

    # AssertionError: assert ['Subset 1', 'test'] == ['test', 'Subset 1']
    assert dm.layer.choices == ['test', 'Subset 1']


def test_data_menu_view_info(specviz_helper, spectrum1d):
    # load 2 data entries
    specviz_helper.load_data(spectrum1d, data_label="test")
    new_spec = specviz_helper.get_spectra(apply_slider_redshift=True)["test"]*0.9
    specviz_helper.load_data(new_spec, data_label="test2")

    dm = specviz_helper.viewers['spectrum-viewer']._obj.glue_viewer.data_menu
    mp = specviz_helper.plugins['Metadata']
    sp = specviz_helper.plugins['Subset Tools']

    sp.import_region(SpectralRegion(6000 * spectrum1d.spectral_axis.unit,
                                    6100 * spectrum1d.spectral_axis.unit),
                     combination_mode='new')
    sp.import_region(SpectralRegion(6200 * spectrum1d.spectral_axis.unit,
                                    6300 * spectrum1d.spectral_axis.unit),
                     combination_mode='new')

    assert dm.layer.choices == ['Subset 2', 'Subset 1', 'test2', 'test']

    dm.layer.selected = ["test2"]
    dm.view_info()
    assert mp.dataset.selected == "test2"

    dm.layer.selected = ["Subset 2"]
    dm.view_info()
    assert sp.subset.selected == "Subset 2"

    dm.layer.selected = ["test", "test2"]
    with pytest.raises(ValueError, match="Only one layer can be selected to view info"):
        dm.view_info()


class TestResizeSubset:
    """
    Test suite for DataMenu resize subset functionality.
    """
    @pytest.fixture(autouse=True)
    def _init(self, imviz_helper, specviz_helper, spectrum1d):
        """
        Autouse fixture to attach the shared helper to the test instance.
        """
        self.imviz_helper = imviz_helper
        self.imviz_helper.load_data(np.zeros((10, 10)), data_label='image', show_in_viewer=True)
        self.imviz_viewer = self.imviz_helper.default_viewer._obj.glue_viewer
        self.imviz_dm = self.imviz_viewer.data_menu
        self.imviz_subset_tools = self.imviz_helper.plugins['Subset Tools']

        self.specviz_helper = specviz_helper
        self.specviz_helper.load(spectrum1d, data_label="Spectrum 1D")
        self.specviz_viewer = self.specviz_helper.app.get_viewer('spectrum-viewer')
        self.specviz_dm = self.specviz_viewer.data_menu
        self.specviz_subset_tools = self.specviz_helper.plugins['Subset Tools']

    def test_allow_resize_subset_in_viewer(self):
        """
        Test _allow_resize_subset_in_viewer with single and composite subsets.
        """
        # Create a circular subset
        self.imviz_viewer.apply_roi(CircularROI(xc=5, yc=5, radius=2))

        # Test without any subsets selected (returns None_
        assert self.imviz_dm._obj._allow_resize_subset_in_viewer() is None
        self.imviz_dm.layer.selected = []
        assert self.imviz_dm._obj._allow_resize_subset_in_viewer() is None

        # Select the subset
        self.imviz_dm.layer.selected = ['Subset 1']

        # Trigger the allow resize check
        self.imviz_dm._obj._allow_resize_subset_in_viewer()

        # Should be enabled for simple ROI
        assert self.imviz_dm._obj.subset_resize_in_viewer_enabled is True

        # Combine to create composite subset
        self.imviz_subset_tools.combination_mode = 'or'
        self.imviz_subset_tools.subset.selected = 'Subset 1'
        self.imviz_viewer.apply_roi(CircularROI(xc=5, yc=5, radius=1))

        # Select the composite subset
        self.imviz_dm.layer.selected = ['Subset 1']

        # Trigger the allow resize check
        self.imviz_dm._obj._allow_resize_subset_in_viewer()

        # Should be disabled for composite ROI
        assert self.imviz_dm._obj.subset_resize_in_viewer_enabled is False

    @pytest.mark.parametrize(
        ("roi", "tool_id"), [
            (CircularROI(xc=5, yc=5, radius=2), 'bqplot:truecircle'),
            (RectangularROI(xmin=2, xmax=7, ymin=3, ymax=8), 'bqplot:rectangle'),
            (EllipticalROI(xc=5, yc=5, radius_x=2, radius_y=3), 'bqplot:ellipse'),
            (CircularAnnulusROI(xc=5, yc=5, inner_radius=1, outer_radius=3), 'bqplot:circannulus')])
    def test_resize_subset_in_viewer_image_rois(self, roi, tool_id):
        """
        Test resize_subset_in_viewer functionality with various image ROIs.
        """
        # Create a circular subset
        self.imviz_viewer.apply_roi(roi)

        # Select the subset for resizing
        self.imviz_dm.layer.selected = ['Subset 1']

        # Resize the subset
        self.imviz_dm.resize_subset_in_viewer()

        # Verify that the correct tool is activated
        assert self.imviz_viewer.toolbar.active_tool_id == tool_id

        # Verify that we're in replace mode
        assert self.imviz_helper.app.session.edit_subset_mode.mode == ReplaceMode

        # Verify that the correct subset is being edited
        subset_grp = [sg for sg in self.imviz_helper.app.data_collection.subset_groups
                      if sg.label == 'Subset 1']
        assert self.imviz_helper.app.session.edit_subset_mode.edit_subset == subset_grp

    @pytest.mark.parametrize('roi', [XRangeROI(min=6000, max=6500),
                                     SpectralRegion(6000 * u.Angstrom, 6500 * u.Angstrom)])
    def test_resize_subset_in_viewer_xrange_roi(self, roi):
        """
        Test resize_subset_in_viewer functionality with an XRange ROI and
        Spectral Region (implicitly) converted to XRange ROI.
        """
        if isinstance(roi, XRangeROI):
            # Create an XRange subset
            self.specviz_viewer.apply_roi(roi)
        elif isinstance(roi, SpectralRegion):
            self.specviz_subset_tools.import_region(roi)

        # Select and resize
        self.specviz_dm.layer.selected = ['Subset 1']
        self.specviz_dm.resize_subset_in_viewer()

        # Verify that the correct tool is activated
        assert self.specviz_viewer.toolbar.active_tool_id == 'bqplot:xrange'
        assert self.specviz_helper.app.session.edit_subset_mode.mode == ReplaceMode

    def test_resize_subset_in_viewer_errors(self):
        """
        Test that resize_subset_in_viewer raises errors where expected.
        """
        # Create two subsets
        self.imviz_viewer.apply_roi(CircularROI(xc=3, yc=3, radius=2))
        self.imviz_subset_tools.combination_mode = 'new'
        self.imviz_viewer.apply_roi(CircularROI(xc=7, yc=7, radius=2))

        # Select both subsets
        self.imviz_dm.layer.selected = ['Subset 1', 'Subset 2']

        # Should raise error
        msg = 'Only one layer can be selected to resize subset.'
        with pytest.raises(ValueError, match=msg):
            self.imviz_dm.resize_subset_in_viewer()

        # Select data layer instead of subset
        self.imviz_dm.layer.selected = ['image']

        # Should raise error
        msg = 'Selected layer is not a subset.'
        with pytest.raises(ValueError, match=msg):
            self.imviz_dm.resize_subset_in_viewer()

        # Select the subset again
        self.imviz_dm.layer.selected = ['Subset 1']

        # Retrieve the subset group and replace its subset_state with a
        # mock object that is not a RoiSubsetState/RangeSubsetState to
        # simulate an unsupported ROI.
        subset_grp = [sg for sg in self.imviz_helper.app.data_collection.subset_groups
                      if sg.label == 'Subset 1']

        # Replace the subset_state with a plain object (unsupported)
        subset_grp[0].subset_state = object()

        # Try to resize and assert the specific error message
        msg = 'Selected subset does not have a supported ROI.'
        with pytest.raises(ValueError, match=msg):
            self.imviz_dm.resize_subset_in_viewer()


def test_catalog_excluded_from_layer_reordering(imviz_helper, image_2d_wcs,
                                                sky_coord_only_source_catalog):
    """
    Test that catalog layers are excluded from zordermanagement in image viewers.

    To do this, load an image, align by WCS, then load a catalog. The catalog
    layer should always be at the top of the layer stack. Then, add a subset to
    ensure the catalog layer remains on top. The addition of a 'Default Orientation'
    and a subset layer will test the reordering functionality of the data menu, and
    ensure that catalog layers are excluded from reordering and always remain on top.
    """

    data = NDData(np.ones((128, 128)), wcs=image_2d_wcs)
    imviz_helper.load(data)

    imviz_helper.plugins['Orientation'].align_by = 'WCS'

    # load catalog
    imviz_helper.app.state.catalogs_in_dc = True
    imviz_helper.load(sky_coord_only_source_catalog, data_label='catalog')

    # load catalog. with the presence of a 'Default Orientation' layer that is
    # NOT listed in the data menu, loading a catalog will cause the layer
    # reordering logic in data_menu to run, and scatter layers should remain unchanged
    layers = imviz_helper.default_viewer.data_menu._obj._viewer.layers
    assert layers[-1].layer.label == 'catalog'

    # Now add a subset and ensure catalog layer remains on top
    subset_tools = imviz_helper.plugins['Subset Tools']
    subset_tools.import_region(CircularROI(xc=0, yc=0, radius=1))
    assert layers[-1].layer.label == 'catalog'



import os
from traitlets import Any, Unicode, observe
from ipywidgets import widget_serialization
from solara import FileBrowser, reactive
import reacton
from pathlib import Path

from jdaviz.core.registries import loader_resolver_registry
from jdaviz.core.loaders.resolvers import BaseResolver
from jdaviz.core.user_api import LoaderUserApi


__all__ = ['FileResolver']


@loader_resolver_registry('file')
class FileResolver(BaseResolver):
    template_file = __file__, "file.vue"
    default_input = 'filepath'
    default_input_cast = str

    file_chooser_widget = Any().tag(sync=True, **widget_serialization)
    filepath = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        # NOTE: file_chooser_dir must always be an absolute path or else its impossible to
        # navigate higher in the directory tree
        self.file_chooser_dir = reactive(Path(os.path.abspath(os.environ.get('JDAVIZ_START_DIR', os.path.curdir))))  # noqa
        self.filepath_reactive = reactive(self.filepath)
        self.file_chooser_widget_el = FileBrowser(directory=self.file_chooser_dir,
                                                  selected=self.filepath_reactive,
                                                  on_path_select=self._on_file_chooser_path_changed,
                                                  can_select=True)
        self.file_chooser_widget, rc = reacton.render(self.file_chooser_widget_el)
        super().__init__(*args, **kwargs)

    @property
    def user_api(self):
        return LoaderUserApi(self, expose=['filepath'])

    @classmethod
    def from_input(cls, app, inp, **kwargs):
        # prevent errors from solara being raised if input is not valid
        if not os.path.exists(inp) or not os.path.isfile(inp):
            raise ValueError(f"'{inp}' is not a valid file path.")
        return super().from_input(app, inp, **kwargs)

    def _on_file_chooser_path_changed(self, path):
        if self.filepath_reactive.value is not None:
            self.filepath = os.path.join(self.file_chooser_dir.value, self.filepath_reactive.value)
        else:
            self.filepath = ''

    @observe('filepath')
    def _on_filepath_changed(self, change):
        # when the filepath traitlet is changed, need to update the
        # file_chooser_widget to match the corresponding path
        if self.filepath == '':
            return
        self.filepath_reactive.value = Path(self.filepath)
        self._resolver_input_updated()
        if not os.path.exists(self.filepath) or not os.path.isfile(self.filepath):
            # consider empty if a directory or non-existent file is selected
            self.parsed_input_is_empty = True

    @property
    def is_valid(self):
        return os.path.exists(self.filepath) and os.path.isfile(self.filepath)  # noqa

    @property
    def default_label(self):
        return os.path.splitext(os.path.basename(self.filepath))[0] if self.filepath else None

    def parse_input(self):
        return self.filepath



from astropy.coordinates import SkyCoord
from astropy import units as u

from pyvo.utils import vocabularies  # noqa: F401
from pyvo import registry
from pyvo.dal.exceptions import DALFormatError, DALQueryError
from requests.exceptions import ConnectionError as RequestConnectionError
from traitlets import Bool, Any, List, observe

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import loader_resolver_registry
from jdaviz.core.template_mixin import (
    SelectPluginComponent,
    with_spinner,
)
from jdaviz.core.loaders.resolvers import BaseConeSearchResolver
from jdaviz.core.user_api import LoaderUserApi

__all__ = ["VOResolver"]


@loader_resolver_registry("virtual observatory")
class VOResolver(BaseConeSearchResolver):
    template_file = __file__, "vo.vue"

    waveband_items = List().tag(sync=True)
    waveband_selected = Any().tag(sync=True)  # Any to accept Nonetype
    resource_filter_coverage = Bool(False).tag(sync=True)
    resource_items = List([]).tag(sync=True)
    resource_selected = Any().tag(sync=True)  # Any to accept Nonetype
    resources_loading = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Waveband properties to filter available registry resources
        self.waveband = SelectPluginComponent(
            self, items="waveband_items", selected="waveband_selected"
        )

        self.waveband.choices = (
            w.lower() for w in vocabularies.get_vocabulary("messenger")["terms"]
        )

        self.waveband_selected = ""

        self._full_registry_results = None
        self.resource_selected = ""
        self.resource = SelectPluginComponent(
            self, items="resource_items", selected="resource_selected"
        )
        self.resource.choices = []

    @property
    def user_api(self):
        return LoaderUserApi(
            self,
            expose=[
                "viewer", "coordframe", "radius", "radius_unit",
                "source",
                "resource_filter_coverage", "waveband", "resource",
                "query_archive"
            ],
        )

    @observe("waveband_selected",
             "source", "coordframe_selected",
             "radius", "radius_unit_selected",
             "resource_filter_coverage")
    @with_spinner(spinner_traitlet="resources_loading")
    def query_registry_resources(self, event={}):
        """
        Query Virtual Observatory registry for all SIA services
        that serve data in that waveband around the source.
        Then update the dropdown accordingly.
        """
        # If waveband was changed to nothing, immediately quit
        # Don't throw an error due to trigger by plugin init
        if not self.waveband_selected:
            return

        # No need to update if the change was from source but coverage filtering is off
        if (event.get("name") in ("source", "coordframe_selected", "radius", "radius_unit_selected")
                and not self.resource_filter_coverage):
            return

        # Can't filter by coverage if we don't have a source to filter on
        if self.resource_filter_coverage and not self.source:
            error_msg = (
                "Source is required for registry querying when coverage filtering is enabled. "
                + (
                    "Please enter your coordinates above "
                    if self.viewer == "Manual"
                    else f"Load data into viewer {self.viewer} first before querying "
                )
                + "or disable coverage filtering."
            )
            self.hub.broadcast(SnackbarMessage(error_msg, sender=self, color="error"))
            raise ValueError(error_msg)

        # Clear existing resources list
        self.resource.choices = []
        self.resource_selected = ""

        try:
            registry_args = [
                registry.Servicetype("sia"),
                registry.Waveband(self.waveband_selected),
            ]
            # If coverage filtering is enabled, lookup current
            # source coordinate and add a Spatial search constraint
            if self.resource_filter_coverage:
                try:
                    # First parse user-provided source as direct coordinates
                    coord = SkyCoord(
                        self.source, unit=u.deg, frame=self.coordframe_selected
                    )
                except Exception:
                    try:
                        # If that didn't work, try parsing it as an object name
                        coord = SkyCoord.from_name(
                            self.source, frame=self.coordframe_selected
                        )
                    except Exception as e:
                        self.hub.broadcast(
                            SnackbarMessage(
                                f"Unable to resolve source coordinates: {self.source}",
                                sender=self,
                                color="error",
                                traceback=e
                            )
                        )
                        raise LookupError(
                            f"Unable to resolve source coordinates: {self.source}"
                        )
                registry_args.append(
                    registry.Spatial(
                        (coord, (self.radius * u.Unit(self.radius_unit.selected))),
                        intersect="overlaps",
                    )
                )
            self._full_registry_results = registry.search(*registry_args)
            self.resource.choices = list(
                self._full_registry_results.getcolumn("short_name")
            )
        except DALFormatError as e:
            if type(e.cause) is RequestConnectionError:
                self.hub.broadcast(
                    SnackbarMessage(
                        f"Can't connect to VO registry. Check your internet connection: {e}",
                        sender=self,
                        color="error",
                        traceback=e
                    )
                )
            else:
                raise e
        except Exception as e:
            self.hub.broadcast(
                SnackbarMessage(
                    f"An error occured querying the VO Registry: {e}",
                    sender=self,
                    color="error",
                    traceback=e
                )
            )
            raise

    @with_spinner(spinner_traitlet="results_loading")
    def query_archive(self):
        """
        Once a specific VO resource is selected, query it with the user-specified source target.
        User input for source is first attempted to be parsed as a SkyCoord coordinate. If not,
        then attempts to parse as a target name.
        """
        try:
            # Query SIA service
            # Service is indexed via short name (resource_selected), which is the suggested way
            # according to PyVO docs. Though disclaimer that collisions COULD occur. If so,
            # consider indexing on the full IVOID, which is guaranteed unique.
            sia_service = self._full_registry_results[
                self.resource_selected
            ].get_service(service_type="sia")
            try:
                # First parse user-provided source as direct coordinates
                coord = SkyCoord(
                    self.source, unit=u.deg, frame=self.coordframe_selected
                )
            except Exception:
                try:
                    # If that didn't work, try parsing it as an object name
                    coord = SkyCoord.from_name(
                        self.source, frame=self.coordframe_selected
                    )
                except Exception as e:
                    self.hub.broadcast(
                        SnackbarMessage(
                            f"Unable to resolve source coordinates: {self.source}",
                            sender=self,
                            color="error",
                            traceback=e
                        )
                    )
            # Once coordinate lookup is complete, search service using these coords.
            try:
                sia_results = sia_service.search(
                    coord,
                    size=(
                        (self.radius * u.Unit(self.radius_unit.selected))
                        if self.radius > 0.0
                        else None
                    ),
                    format="image/fits",
                )
            except DALQueryError as e:
                # We've run into issues where the service assumes a FORMAT and injects it for us.
                # If the "image/fits" is duplicated, remove our requested format and rely on theirs
                if "Wrong FORMAT=image/fits,image/fits" in str(e):
                    sia_results = sia_service.search(
                        coord,
                        size=(
                            (self.radius * u.Unit(self.radius_unit.selected))
                            if self.radius > 0.0
                            else None
                        ),
                    )
                else:
                    self.hub.broadcast(
                        SnackbarMessage(
                            f"Query failed: {e}",
                            sender=self,
                            traceback=e,
                            color="error",
                        )
                    )
            if len(sia_results) == 0:
                self.hub.broadcast(
                    SnackbarMessage(
                        f"No observations returned at coords {coord} from VO SIA resource: {sia_service.baseurl}",  # noqa: E501
                        sender=self,
                        color="error",
                    )
                )
            else:
                self.hub.broadcast(
                    SnackbarMessage(
                        f"{len(sia_results)} SIA results found!",
                        sender=self,
                        color="success",
                    )
                )
        except Exception as e:
            self.hub.broadcast(
                SnackbarMessage(
                    f"Unable to locate files for source {self.source}: {e}",
                    sender=self,
                    color="error",
                    traceback=e
                )
            )
        try:
            self._output = sia_results.to_table()
        except Exception as e:
            self.hub.broadcast(
                SnackbarMessage(
                    f"Unable to populate table for source {self.source}: {e}",
                    sender=self,
                    color="error",
                    traceback=e
                )
            )
        self._resolver_input_updated()

    def vue_query_archive(self, _=None):
        self.query_archive()

    def parse_input(self):
        return self._output



import numpy as np
import pytest
from pathlib import Path
from itertools import product

from astropy import units as u
from astropy.table import Table
from astropy.io import fits
from astropy.wcs import WCS
from astroquery.mast import Mast, MastMissions
from gwcs import WCS as GWCS
from specutils import SpectralRegion, Spectrum

from jdaviz.core.registries import loader_resolver_registry
from jdaviz.core.loaders.resolvers import find_matching_resolver
from jdaviz.utils import cached_uri


def test_loaders_registry(specviz_helper):
    # built-in loaders: file, url, etc
    assert len(specviz_helper.loaders) == len(loader_resolver_registry.members)


def test_open_close(specviz_helper):
    specviz_helper.app.state.dev_loaders = True

    assert specviz_helper.app.state.drawer_content == ''

    loader = specviz_helper.loaders['file']
    loader.open_in_tray()
    assert specviz_helper.app.state.drawer_content == 'loaders'
    loader.close_in_tray()
    assert specviz_helper.app.state.drawer_content == 'loaders'
    loader.close_in_tray(close_sidebar=True)
    assert specviz_helper.app.state.drawer_content == ''

    subset_plg = specviz_helper.plugins['Subset Tools']
    subset_plg.open_in_tray()
    assert specviz_helper.app.state.drawer_content == 'plugins'

    assert subset_plg._obj.loader_panel_ind is None  # loader panel not open
    subset_plg._obj.loaders['url'].open_in_tray()
    assert subset_plg._obj.loader_panel_ind == 0  # loader panel open
    subset_plg._obj.loaders['url'].close_in_tray()
    assert subset_plg._obj.loader_panel_ind is None  # loader panel not open
    subset_plg._obj.loaders['url'].close_in_tray(close_sidebar=True)
    assert specviz_helper.app.state.drawer_content == ''


def test_resolver_matching(specviz_helper):
    sp = Spectrum(spectral_axis=np.array([1, 2, 3])*u.nm,
                  flux=np.array([1, 2, 3])*u.Jy)

    res_sp = find_matching_resolver(specviz_helper.app, sp)
    assert res_sp._obj._registry_label == 'object'
    assert res_sp.format == '1D Spectrum'

    specviz_helper._load(sp)
    assert len(specviz_helper.app.data_collection) == 1


def test_dbg_access(deconfigged_helper):
    test_data = np.array([1, 2, 3])
    deconfigged_helper.loaders['object'].object = test_data

    # test that _get_loader is successful to grab the object
    # and expose the underlying error
    with pytest.raises(ValueError, match="Cannot load as image with ndim=1"):  # noqa
        deconfigged_helper._get_loader('object', 'object', 'Image')


def test_trace_importer(specviz2d_helper, spectrum2d):
    specviz2d_helper._load(spectrum2d, format='2D Spectrum')

    trace = specviz2d_helper.plugins['2D Spectral Extraction'].export_trace()

    res_sp = find_matching_resolver(specviz2d_helper.app, trace)
    assert res_sp._obj._registry_label == 'object'
    assert res_sp.format == 'Trace'

    # import through loader API
    ldr = specviz2d_helper.loaders['object']
    ldr.object = trace
    assert ldr.format == 'Trace'
    ldr.importer.data_label = 'Trace 1'
    ldr.load()
    assert specviz2d_helper.app.data_collection[-1].label == 'Trace 1'

    # import through load method
    specviz2d_helper._load(trace, data_label='Trace 2')
    assert specviz2d_helper.app.data_collection[-1].label == 'Trace 2'


def test_spectrum2d_viewer_options(deconfigged_helper, spectrum2d):
    ldr = deconfigged_helper.loaders['object']
    ldr.object = spectrum2d

    assert ldr.importer.viewer.create_new == '2D Spectrum'
    assert ldr.importer.viewer.new_label == '2D Spectrum'
    assert ldr.importer.ext_viewer.create_new == '1D Spectrum'
    assert ldr.importer.ext_viewer.new_label == '1D Spectrum'

    ldr.importer.ext_viewer.create_new = ''
    assert ldr.importer.ext_viewer == []

    assert ldr.importer.auto_extract
    ldr.load()

    # created 2D Spectrum viewer, did auto-extract,
    # but did not create 1D Spectrum viewer
    assert len(deconfigged_helper.viewers) == 1
    assert len(deconfigged_helper.app.data_collection) == 2


def test_markers_specviz2d_unit_conversion(specviz2d_helper, spectrum2d):
    data = np.zeros((5, 10))
    data[3] = np.arange(10)
    spectrum2d = Spectrum(flux=data*u.MJy, spectral_axis=data[3]*u.AA)
    specviz2d_helper.load_data(spectrum2d)


@pytest.mark.remote_data
@pytest.mark.filterwarnings(r"ignore::astropy.wcs.wcs.FITSFixedWarning")
def test_fits_spectrum2d(deconfigged_helper):
    uri = cached_uri('mast:jwst/product/jw02123-o001_v000000353_nirspec_f170lp-g235h_s2d.fits')
    if 'mast' in uri:
        ldr = deconfigged_helper.loaders['url']
        ldr.cache = True
        ldr.url = uri
    else:
        ldr = deconfigged_helper.loaders['file']
        ldr.filepath = uri

    # Default format may not be 2D Spectrum
    assert 'Image' in ldr.format.choices
    assert '2D Spectrum' in ldr.format.choices
    ldr.format = '2D Spectrum'
    assert ldr.importer._obj._parser.__class__.__name__ == 'FITSParser'

    ldr.load()

    # ensure get_data works, retrieves a Spectrum1D object, and has spectral WCS attached correctly
    sp2d = deconfigged_helper.get_data('jw02123-o001_v000000353_nirspec_f170lp-g235h_s2d')  # noqa
    assert isinstance(sp2d, Spectrum)
    assert str(sp2d.spectral_axis.unit) == 'um'

    sp1d = deconfigged_helper.get_data('jw02123-o001_v000000353_nirspec_f170lp-g235h_s2d (auto-ext)')  # noqa
    assert isinstance(sp1d, Spectrum)
    assert str(sp1d.spectral_axis.unit) == 'um'


@pytest.mark.remote_data
@pytest.mark.filterwarnings(r"ignore::astropy.wcs.wcs.FITSFixedWarning")
@pytest.mark.filterwarnings(r"ignore::asdf.exceptions.AsdfPackageVersionWarning")
def test_jwst_wfss_bsub(deconfigged_helper):
    uri = 'https://stsci.box.com/shared/static/k5l446jid348jk343m8r4vvbkzk4syom.fits'
    ldr = deconfigged_helper.loaders['url']
    ldr.cache = True
    ldr.url = uri

    assert ldr.format == '2D Spectrum'

    ldr.load()

    sp1d = deconfigged_helper.get_data('k5l446jid348jk343m8r4vvbkzk4syom (auto-ext)')  # noqa
    assert isinstance(sp1d, Spectrum)
    assert str(sp1d.spectral_axis.unit) == 'pix'


# TODO: Skip until file is uploaded to box
@pytest.mark.skip
def test_fits_spectrum_list_L3_wfss(deconfigged_helper):
    ldr = deconfigged_helper.loaders['url']
    # placeholder for WFSS L3 URI
    ldr.url = None

    # ldr = deconfigged_helper.loaders['file']
    # ldr.filepath = './jdaviz/notebooks/WFSS_fits/jw01076-o103_t0000_nircam_f356w-grismr_x1d.fits'  # noqa
    ldr.format = '1D Spectrum List'

    # 1_117 is completely masked
    sources_obj = ldr.importer.sources
    number_combos = product((1, 2), (9, 17, 23))
    sources_obj.selected = [f'Exposure {e_num}, Source ID: {s_id}'
                            for e_num, s_id in number_combos]
    ldr.load()

    assert len(deconfigged_helper.data_labels) == len(sources_obj.selected)
    dc = deconfigged_helper.app.data_collection
    assert len(dc) == len(sources_obj.selected)
    assert len(deconfigged_helper.viewers) == 1

    filestem = Path(ldr.filepath).stem
    for e_num, s_id in number_combos:
        spec = deconfigged_helper.get_data(f'{filestem}_EXP-{e_num}_ID-{s_id}')
        assert isinstance(spec, Spectrum)
        assert str(spec.spectral_axis.unit) == 'um'


@pytest.mark.remote_data
def test_resolver_url(deconfigged_helper, fake_classes_in_registries):

    def ignore_custom_loaders(existing_choices):
        return [choice for choice in existing_choices if choice not in fake_classes_in_registries]

    loader = deconfigged_helper.loaders['url']

    # no url, no valid formats
    assert len(loader.format.choices) == 0

    # non-valid input
    loader.url = 'not-valid-url'
    assert len(loader.format.choices) == 0

    # s3 input
    loader.url = "s3://stpubdata/jwst/public/jw02727/L3/t/o002/jw02727-o002_t062_nircam_clear-f277w_i2d.fits"  # noqa: E501
    assert loader._obj.url_scheme == 's3'
    assert len(loader.format.choices) > 0

    # https valid input
    loader.url = 'https://stsci.box.com/shared/static/exnkul627fcuhy5akf2gswytud5tazmw.fits'  # noqa

    # may change with future importers
    assert len(ignore_custom_loaders(loader.format.choices)) == 4
    assert loader.format.selected == 'Image'  # default may change with future importers

    # test target filtering
    assert len(loader.target.choices) > 1
    assert loader.target.selected == 'Any'
    loader.target = '1D Spectrum'

    # may change with future importers
    assert len(ignore_custom_loaders(loader.format.choices)) == 2
    assert loader.format == '1D Spectrum List'  # default may change with future importers
    assert loader.importer.data_label == 'exnkul627fcuhy5akf2gswytud5tazmw'  # noqa

    loader.target = 'Any'
    assert len(ignore_custom_loaders(loader.format.choices)) == 4
    loader.format = '2D Spectrum'
    assert loader.importer.data_label == 'exnkul627fcuhy5akf2gswytud5tazmw'  # noqa

    assert len(deconfigged_helper.app.data_collection) == 0
    assert len(deconfigged_helper.viewers) == 0

    loader.load()

    # 2D spectrum and auto-extracted 1D spectrum
    assert len(deconfigged_helper.app.data_collection) == 2
    assert len(deconfigged_helper.viewers) == 2

    with pytest.raises(ValueError, match="Failed query for URI"):
        # NOTE: this test will attempt to reach out to MAST via astroquery
        # even if cache is available.
        deconfigged_helper.load('mast:invalid')


def test_resolver_table_as_query(deconfigged_helper):
    ldr = deconfigged_helper.loaders['object']

    obs_table = Table({'id': [1, 2, 3], 'fileSetName': ['a', 'b', 'c']})
    file_table = Table({'id': [1, 2, 3], 'url': ['a', 'b', 'c']})
    invalid_table = Table({'id': [1, 2, 3], 'name': ['a', 'b', 'c']})

    assert ldr._obj.parsed_input_is_query is False

    ldr.object = obs_table
    assert ldr._obj.parsed_input_is_query is True
    assert ldr._obj.observation_table_populated is True
    assert ldr._obj.file_table_populated is False

    ldr.object = file_table
    assert ldr._obj.parsed_input_is_query is True
    assert ldr._obj.observation_table_populated is False
    assert ldr._obj.file_table_populated is True

    ldr.object = invalid_table
    assert ldr._obj.parsed_input_is_query is False
    assert ldr._obj.observation_table_populated is False
    assert ldr._obj.file_table_populated is False


@pytest.mark.filterwarnings("ignore::UserWarning")
@pytest.mark.remote_data
def test_resolver_table_as_query_astroquery(deconfigged_helper, tmp_path):
    ldr = deconfigged_helper.loaders['object']

    mast = Mast()
    coords = mast.resolve_object("M101")

    jwst = MastMissions(mission='JWST')

    datasets = jwst.query_region(coords,
                                 radius=3,
                                 select_cols=["sci_stop_time", "sci_targname",
                                              "sci_start_time", "sci_status"],
                                 sort_by=['sci_targname'])

    ldr.object = datasets
    assert ldr._obj.parsed_input_is_query is True
    assert ldr._obj.observation_table_populated is True
    assert ldr._obj.file_table_populated is False
    assert len(ldr.format.choices) == 0

    ldr.observation_table.select_rows(0)

    assert ldr._obj.file_table_populated is True
    assert ldr._obj.get_selected_url() is None

    ldr.file_table.select_rows(0)
    assert ldr._obj.get_selected_url() is not None
    # we can't guarantee any specific format choices since
    # this depends on what data is returned from MAST
    # but let's at least make sure the download was successful
    # and points to a local temporary cache file
    assert len(ldr._obj.output) > 0


def test_invoke_from_plugin(specviz_helper, spectrum1d, tmp_path):
    s = SpectralRegion(5*u.um, 6*u.um)
    local_path = str(tmp_path / 'spectral_region.ecsv')
    s.write(local_path)

    specviz_helper.load_data(spectrum1d)

    loader = specviz_helper.plugins['Subset Tools'].loaders['file']

    assert len(loader.format.choices) == 0
    loader.filepath = local_path
    assert len(loader.format.choices) > 0

    loader.load()


@pytest.mark.parametrize('order', ([0, 1, 2], [0, 2, 1], [2, 0, 1], [2, 1, 0], [1, 2, 0]))
def test_mult_data_types(deconfigged_helper, image_nddata_wcs, spectrum2d, spectrum1d, order):
    datas = [image_nddata_wcs, spectrum2d, spectrum1d]
    load_kwargs = [{'format': 'Image'},
                   {'format': '2D Spectrum', 'auto_extract': True},
                   {'format': '1D Spectrum'}]

    for i in order:
        deconfigged_helper.load(datas[i], **load_kwargs[i])

    # NOTE: 2D Spectrum will also result in auto-extracted 1D Spectrum
    assert len(deconfigged_helper.app.data_collection) == 4
    assert len(deconfigged_helper.viewers) == 3


def test_freq_wavelength_linking(deconfigged_helper, spectrum1d):
    deconfigged_helper.load(spectrum1d, format='1D Spectrum', data_label='sp_wavelength')
    sp1d_freq = Spectrum(spectral_axis=spectrum1d.spectral_axis.to(u.Hz, equivalencies=u.spectral()),  # noqa
                         flux=spectrum1d.flux,
                         uncertainty=spectrum1d.uncertainty,
                         mask=spectrum1d.mask,
                         meta=spectrum1d.meta)
    deconfigged_helper.load(sp1d_freq, format='1D Spectrum', data_label='sp_frequency')

    # flux <> flux, uncertainty <> uncertainty, wavelength <> freq, Pixel 0[x] <> Pixel 1[x]
    assert len(deconfigged_helper.app.data_collection.external_links) == 4


def test_load_image_mult_sci_extension(imviz_helper):
    # test loading an image with multiple SCI extensions and
    # ensure that automatic parenting logic is handled correctly
    arr = np.zeros((2, 2), dtype=np.float32)
    hdul = fits.HDUList([fits.PrimaryHDU(),
                        fits.ImageHDU(arr, name='SCI', ver=1),
                        fits.ImageHDU(arr, name='ERR', ver=1),
                        fits.ImageHDU(arr, name='SCI', ver=2),
                        fits.ImageHDU(arr, name='ERR', ver=2)
                         ])

    # imviz_helper._load(hdul, extension=('SCI,1', 'SCI,2', 'ERR,2'))
    imviz_helper.load_data(hdul, ext=('SCI,1', 'SCI,2', 'ERR,2'))

    assert len(imviz_helper.app.data_collection) == 3
    assert [d.label for d in imviz_helper.app.data_collection] == ['Image[SCI,1]', 'Image[SCI,2]', 'Image[ERR,2]']  # noqa

    assert imviz_helper.app._get_assoc_data_children('Image[SCI,1]') == []
    assert imviz_helper.app._get_assoc_data_children('Image[SCI,2]') == ['Image[ERR,2]']
    assert imviz_helper.app._get_assoc_data_parent('Image[ERR,2]') == 'Image[SCI,2]'


def test_loaders_extension_select(imviz_helper):
    # tests internal logic of SelectFileExtensionComponent
    arr = np.zeros((2, 2), dtype=np.float32)
    hdul = fits.HDUList([fits.PrimaryHDU(),
                        fits.ImageHDU(arr, name='SCI', ver=1),
                        fits.ImageHDU(arr, name='ERR', ver=1),
                        fits.ImageHDU(arr, name='SCI', ver=2),
                        fits.ImageHDU(arr, name='ERR', ver=2)
                         ])

    ldr = imviz_helper.loaders['object']
    ldr.object = hdul

    # test setting extension by index: [name, ver], [name,ver], and name,ver
    ldr.importer.extension = ['1: [SCI,1]', '[SCI,2]', 'ERR,2']
    assert ldr.importer.extension.selected == ['1: [SCI,1]',
                                               '3: [SCI,2]',
                                               '4: [ERR,2]']
    assert ldr.importer.extension.selected_index == [1, 3, 4]

    # if not providing ver, first match will be used
    ldr.importer.extension = 'SCI'
    assert ldr.importer.extension.selected == ['1: [SCI,1]']

    # case of providing index
    ldr.importer.extension = [1, 3]
    assert ldr.importer.extension.selected == ['1: [SCI,1]', '3: [SCI,2]']


def test_load_image_align_by(deconfigged_helper, image_nddata_wcs):
    ldr = deconfigged_helper.loaders['object']
    ldr.object = image_nddata_wcs
    assert 'Image' in ldr.format.choices
    ldr.format = 'Image'

    assert ldr.importer.align_by == 'Pixels'

    ldr.load()

    assert deconfigged_helper.plugins['Orientation'].align_by.selected == 'Pixels'

    ldr.importer.align_by = 'WCS'
    ldr.load()

    assert deconfigged_helper.plugins['Orientation'].align_by.selected == 'WCS'


@pytest.mark.remote_data
@pytest.mark.parametrize(
    ('gwcs_to_fits_sip', 'expected_cls'),
    ((True, WCS), (False, GWCS),), ids=('True-WCS', 'False-GWCS'))
@pytest.mark.filterwarnings("ignore:Some non-standard WCS keywords were excluded")
def test_gwcs_to_fits_sip(gwcs_to_fits_sip, expected_cls, deconfigged_helper):
    """Test gwcs_to_fits_sip through the importer API."""
    ldr = deconfigged_helper.loaders['url']
    ldr.url = 'https://data.science.stsci.edu/redirect/JWST/jwst-data_analysis_tools/imviz_test_data/jw00042001001_01101_00001_nrcb5_cal.fits'  # noqa
    ldr.importer.gwcs_to_fits_sip = gwcs_to_fits_sip

    ldr.load()

    data = deconfigged_helper.app.data_collection[0]
    assert isinstance(data.coords, expected_cls)


@pytest.mark.remote_data
@pytest.mark.parametrize('helper', ['deconfigged_helper', 'specviz_helper'])
def test_roman_1d_spectrum(helper, request):
    helper = request.getfixturevalue(helper)
    ldr = helper.loaders['url']
    # wfi_spec_combined_1d_r0000201001001001001_0002_WFI01.asdf
    ldr.url = 'https://stsci.box.com/shared/static/rgasl942so9hno2rq9f1xgdeolav59o5.asdf'
    assert len(ldr.importer.extension.choices) > 1
    ldr.format = '1D Spectrum'

    ldr.load()
    assert len(helper.app.data_collection) == 1


@pytest.mark.remote_data
@pytest.mark.parametrize('helper', ['deconfigged_helper', 'specviz2d_helper'])
def test_roman_2d_spectrum(helper, request):
    helper = request.getfixturevalue(helper)
    ldr = helper.loaders['url']
    # wfi_spec_decontaminated_2d_r0000201001001001001_0002_WFI01.asdf
    ldr.url = 'https://stsci.box.com/shared/static/g8yb9hxguy3aedveef9su67lesgd8c6w.asdf'
    assert len(ldr.importer.extension.choices) > 1
    ldr.format = '2D Spectrum'

    ldr.load()
    # 2D spectrum and auto-extracted 1D spectrum
    assert len(helper.app.data_collection) == 2



import os
import warnings
import numpy as np
import threading

import pytest
from astropy.io import fits
from astropy.units.quantity import Quantity

from jdaviz.utils import (alpha_index, download_uri_to_path,
                          get_cloud_fits, cached_uri, escape_brackets,
                          has_wildcard, wildcard_match, _clean_data_for_hash,
                          create_data_hash, parallelize_calculation)

from jdaviz.conftest import FakeSpectrumListImporter


@pytest.mark.parametrize("test_input,expected", [(0, 'a'), (1, 'b'), (25, 'z'), (26, 'aa'),
                                                 (701, 'zz'), (702, '{a')])
def test_alpha_index(test_input, expected):
    assert alpha_index(test_input) == expected


def test_alpha_index_exceptions():
    with pytest.raises(TypeError, match="index must be an integer"):
        alpha_index(4.2)
    with pytest.raises(ValueError, match="index must be positive"):
        alpha_index(-1)


def test_uri_to_download_bad_scheme(imviz_helper):
    uri = "file://path/to/file.fits"
    with pytest.raises(ValueError, match="no valid loaders found for input"):
        imviz_helper.load_data(uri)


@pytest.mark.remote_data
def test_uri_to_download_nonexistent_mast_file(imviz_helper):
    # this validates as a mast uri but doesn't actually exist on mast:
    uri = "mast:JWST/product/jw00000-no-file-here.fits"
    with pytest.raises(ValueError, match='Failed query for URI'):
        # NOTE: this test will attempt to reach out to MAST via astroquery
        # even if cache is available.
        imviz_helper.load_data(uri, cache=False)


@pytest.mark.remote_data
def test_url_to_download_imviz_local_path_warning(imviz_helper):
    url = "https://www.astropy.org/astropy-data/tutorials/FITS-images/HorseHead.fits"

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        imviz_helper.load_data(url, cache=True, local_path='horsehead.fits')


def test_uri_to_download_specviz_local_path_check():
    # NOTE: do not use cached_uri here since no download should occur
    uri = "mast:JWST/product/jw02732-c1001_t004_miri_ch1-short_x1d.fits"
    local_path = download_uri_to_path(uri, cache=False, dryrun=True)  # No download

    # Wrong: '.\\JWST/product/jw02732-c1001_t004_miri_ch1-short_x1d.fits'
    # Correct:  '.\\jw02732-c1001_t004_miri_ch1-short_x1d.fits'
    assert local_path == os.path.join(os.curdir, "jw02732-c1001_t004_miri_ch1-short_x1d.fits")  # noqa: E501


@pytest.mark.remote_data
def test_uri_to_download_specviz(specviz_helper):
    uri = cached_uri("mast:JWST/product/jw02732-c1001_t004_miri_ch1-short_x1d.fits")
    specviz_helper.load_data(uri, cache=True)


@pytest.mark.remote_data
def test_uri_to_download_specviz2d(specviz2d_helper):
    uri = cached_uri("mast:jwst/product/jw01538-o161_t002-s000000001_nirspec_f290lp-g395h-s1600a1_s2d.fits")  # noqa: E501
    specviz2d_helper.load_data(uri, cache=True)


@pytest.mark.remote_data
def test_load_s3_fits(imviz_helper):
    """Test loading a JWST FITS file from an S3 URI into Imviz."""
    s3_uri = "s3://stpubdata/jwst/public/jw02727/L3/t/o002/jw02727-o002_t062_nircam_clear-f277w_i2d.fits"  # noqa: E501
    imviz_helper.load_data(s3_uri)
    assert len(imviz_helper.app.data_collection) > 0


@pytest.mark.remote_data
def test_get_cloud_fits_ext():
    s3_uri = "s3://stpubdata/jwst/public/jw02727/L3/t/o002/jw02727-o002_t062_nircam_clear-f277w_i2d.fits"  # noqa: E501
    hdul = get_cloud_fits(s3_uri)
    assert isinstance(hdul, fits.HDUList)

    hdul = get_cloud_fits(s3_uri, ext="SCI")
    assert isinstance(hdul, fits.HDUList)

    hdul = get_cloud_fits(s3_uri, ext=["SCI"])
    assert isinstance(hdul, fits.HDUList)


class FakeObject:
    def __init__(self):
        pass


def test_escape_brackets():
    # Test that wildcard_match escapes brackets properly
    assert escape_brackets("") == ""
    assert escape_brackets("no brackets") == "no brackets"
    assert escape_brackets("some [brackets]") == "some [[]brackets[]]"
    assert escape_brackets("s[o]me [br]a[]ckets]") == "s[[]o[]]me [[]br[]]a[[][]]ckets[]]"


def test_has_wildcard():
    assert has_wildcard("") is False
    assert has_wildcard("no wildcards") is False
    assert has_wildcard("some * wildcard") is True
    assert has_wildcard("some ? wildcard") is True
    # Don't check for brackets anymore, just * and ?
    assert has_wildcard("some [brackets] wildcard") is False
    assert has_wildcard("some [brackets] and * wildcard") is True
    assert has_wildcard("some [brackets] and ? wildcard") is True
    assert has_wildcard("*") is True
    assert has_wildcard("?") is True


def test_wildcard_match_basic(deconfigged_helper, premade_spectrum_list):
    default_choices = ['some choice', 'some good choice', 'good choice', 'maybe a bad choice']
    test_obj = FakeObject()

    # No choices in obj or provided (choices is None)
    match_result = wildcard_match(test_obj, '*')
    assert match_result == '*'

    # No choices in obj or provided (choices is not None)
    match_result = wildcard_match(test_obj, '*', choices=[])
    assert match_result == ''

    # No choices in obj or provided, but multiselect is an attribute and false,
    # not all wildcards are *
    test_obj.multiselect = False
    match_result = wildcard_match(test_obj, '?*', choices=[])
    assert match_result == '?*'

    # Multiselect is not an attribute yet so we retain the value of the input string
    match_result = wildcard_match(test_obj, '*', choices=default_choices[0])
    assert match_result == ''

    test_obj.allow_multiselect = True
    # No choices in obj or provided, but multiselect is an attribute
    # and set to True inside the function
    match_result = wildcard_match(test_obj, '*', choices=[])
    assert match_result == []

    # No choices in obj or provided, but multiselect is an attribute
    # and set to True inside the function, value is a list of * wildcards
    match_result = wildcard_match(test_obj, ['*', '*'], choices=[])
    assert match_result == []

    # No choices in obj or provided, but multiselect is an attribute
    # and set to True inside the function, value list contains non-* wildcards
    match_result = wildcard_match(test_obj, ['*', '?', '*?'], choices=[])
    assert match_result == ['*', '?', '*?']

    # Add choices attribute, no matches
    test_obj.choices = default_choices
    match_result = wildcard_match(test_obj, 'not good*')
    assert match_result == ['not good*']

    # Test that the function kwarg `choices` overrides the object's choices attribute
    match_result = wildcard_match(test_obj, 'bad*', choices=default_choices[:-1])
    assert match_result == ['bad*']

    test_selections = {
        # Test all
        '*': default_choices,
        # Test no wildcard + wildcard
        ('some', 'some*'): ['some'] + default_choices[:2],
        # Test repeats
        ('*', '* good *'): default_choices,
        # Test single selection
        'some*': default_choices[:2],
        # Test multi-wildcard
        ('*', 'good*'): default_choices,
        # Test multi-selection
        ('some*', 'good*'): default_choices[:-1]}

    for selection, expected in test_selections.items():
        # Set and reset
        test_obj.multiselect = False
        match_result = wildcard_match(test_obj, selection)
        assert test_obj.multiselect is True
        assert match_result == expected

    # Making sure a stand-in for a SelectPluginComponent object with an attribute
    # that has `choices` works as expected
    fake_importer = FakeSpectrumListImporter(app=deconfigged_helper.app,
                                             resolver=deconfigged_helper.loaders['object']._obj,
                                             parser=None,
                                             input=premade_spectrum_list)
    test_obj = fake_importer.sources

    """
    Left here for reference, premade_spectrum_list has 5 spectra:
    default_choices = ['1D Spectrum at index: 0',
                       '1D Spectrum at index: 1',
                       'Exposure 0, Source ID: 0000',
                       'Exposure 0, Source ID: 1111',
                       'Exposure 1, Source ID: 1111']
    """

    test_selections = {
        # Test all
        '*': test_obj.choices,
        # Test repeats
        ('*', '*:*'): test_obj.choices,
        # Test single selection
        '1D Spectrum at index: ?': test_obj.choices[:2],
        # Test multi-wildcard
        '*Exposure*': test_obj.choices[2:],
        # Test multi-selection
        ('*at index: 1', 'Exposure 0*'): test_obj.choices[1:-1]}

    for selection, expected in test_selections.items():
        # Reset
        test_obj.multiselect = False
        match_result = wildcard_match(test_obj, selection)
        assert test_obj.multiselect is True
        assert match_result == expected


@pytest.mark.parametrize('n_cpu', [1, 2, None])
class TestParallelizeCalculation:
    """
    Test for parallelize_calculation in jdaviz.utils.

    This module provides a simple embarrassingly parallel test: compute
    squares of integers using no-argument worker callables, then collect
    results via a thread-safe callback. The test asserts all expected
    results are collected.
    """
    values = range(10)

    def test_parallelize_calculation_squares(self, n_cpu):
        """
        Verify that parallelize_calculation runs a set of no-arg worker
        callables in parallel and that the collect_result_callback receives
        each result.

        The test uses a small deterministic set of integers and checks that
        the collected results equal the expected squares.
        """
        # Create worker callables that return the square of each value.
        workers = [lambda v=v: v * v for v in self.values]

        collected = []
        lock = threading.Lock()

        def collect_result_callback(res):
            """
            Thread-safe collector that appends results to the shared list.
            Using a lock makes the collector deterministic and robust across
            backends and Python implementations, preventing flaky test failures,
            but it's not strictly necessary in the Jdaviz context (for now).
            """
            with lock:
                collected.append(res)

        # Use the default number of CPUs for the test to exercise parallel execution.
        parallelize_calculation(workers, collect_result_callback, n_cpu=n_cpu)

        # After execution, collected should contain exactly the squares, in
        # some order.
        assert sorted(collected) == sorted([v * v for v in self.values])

    def test_callback_called_once_per_worker(self, n_cpu):
        """
        Ensure the collect_result_callback is invoked exactly once per
        worker. The callback records how many times each result value is
        seen; after running, each expected result must have been seen once.

        This will fail if the callback is called multiple times per worker
        as might be the case in a concurrent execution environment should
        something change in the future.
        """
        # Workers return unique values (their index) so we can count calls
        workers = [lambda v=v: v for v in self.values]

        counts = {}
        lock = threading.Lock()

        def collect_result_callback(res):
            """
            Record the number of times a particular result value is seen.
            """
            with lock:
                counts[res] = counts.get(res, 0) + 1

        parallelize_calculation(workers, collect_result_callback, n_cpu=n_cpu)

        # Total callbacks must equal number of workers and each value seen once.
        assert sum(counts.values()) == len(workers)
        assert set(counts.keys()) == set(self.values)
        assert all(c == 1 for c in counts.values())

    def test_exception_in_worker_propagates(self, n_cpu):
        """
        Ensure that an exception raised in a worker propagates out of
        `parallelize_calculation` regardless of worker sequence and that
        since collection happens afterwards, no results are collected.
        """
        bad_worker_msg = 'bad worker :('
        good_worker_msg = 'good worker :)'

        def bad_worker():
            raise ValueError(bad_worker_msg)

        # A worker that would succeed if it were run.
        def good_worker():
            return good_worker_msg

        workers1 = [bad_worker, good_worker]
        workers2 = [good_worker, bad_worker]
        workers3 = [good_worker, bad_worker, bad_worker]

        collected = []

        def collect_result_callback(res):
            # If this is called, append to collected so we can detect it.
            collected.append(res)

        # Since we do collection after the fact, we expect the exception
        # regardless of the worker order.
        for workers in (workers1, workers2, workers3):
            with pytest.raises(ValueError) as err:
                parallelize_calculation(workers, collect_result_callback, n_cpu=n_cpu)

            # The raised exception should contain our message.
            assert bad_worker_msg in str(err.value)

            # Because the bad worker raised the collect callback should
            # not be called at all.
            assert collected == []


@pytest.mark.parametrize('input_data',
                         [np.arange(10), '12345',
                          np.ma.masked_array(np.arange(10), mask=[0, 1, 0, 0, 1, 1, 0, 0, 1, 0]),
                          Quantity(np.arange(10), 'm/s'),
                          'spectrum1d', 'partially_masked_spectrum1d',
                          'spectrum2d', 'image_hdu_nowcs', 'image_nddata_wcs'])
class TestDataHash:
    def test_clean_data_for_hash(self, input_data, request):
        data = input_data
        if isinstance(input_data, str) and not input_data.isnumeric():
            data = request.getfixturevalue(input_data)

        result_arr, result_mask_arr, result_unit_str = _clean_data_for_hash(data)

        # Main array must always be an ndarray
        assert isinstance(result_arr, np.ndarray)

        # Determine expected unit string from possible container types.
        expected_unit = None
        # Try for a mask while we're at it
        expected_mask = getattr(data, 'mask', None)
        if hasattr(data, 'unit'):
            expected_unit = str(data.unit)
        elif hasattr(data, 'flux') and hasattr(data.flux, 'unit'):
            expected_unit = str(data.flux.unit)
            # Spectrum objects always have a mask
        elif hasattr(data, 'data') and hasattr(data.data, 'unit'):
            expected_unit = str(data.data.unit)
            expected_mask = expected_mask or getattr(data.data, 'mask', None)

        assert expected_unit == result_unit_str
        if expected_mask is not None:
            assert np.allclose(expected_mask, result_mask_arr)
        else:
            assert result_mask_arr is None

    def test_create_data_hash(self, input_data, request):
        """Run create_data_hash on several existing test fixtures and
        ensure determinism (identical input -> identical hash).
        """
        data = input_data
        if isinstance(input_data, str) and not input_data.isnumeric():
            data = request.getfixturevalue(input_data)

        h1 = create_data_hash(data)
        h2 = create_data_hash(data)
        # Determinism: repeated calls should be equal (can be None)
        assert isinstance(h1, str)
        try:
            assert h1 == h2
        except ValueError:
            assert np.allclose(h1, h2)

        try:
            data += data
        except TypeError:
            pass
        else:
            h3 = create_data_hash(data)
            # Changing the data should change the hash
            try:
                assert h1 != h3
            except ValueError:
                assert not np.allclose(h1, h3)


def test_create_data_hash_none():
    """Checks behavior for None and unsupported types."""
    assert create_data_hash(None) is None
    assert create_data_hash([]) is None
    assert create_data_hash([None, None, None]) is None
    assert create_data_hash(np.array([])) is None
    assert create_data_hash(np.array([None, None, None])) is None



from echo import delay_callback, ignore_callback
import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord, Angle
from astropy.wcs import WCS
from gwcs.wcs import WCS as GWCS

from jdaviz.utils import get_top_layer_index
from jdaviz.configs.imviz.wcs_utils import get_compass_info


class AID:
    """
    Common API methods for image viewers in astronomy, called
    the Astro Image Display API (AIDA)[1]_.

    References
    ----------
    .. [1] https://github.com/astropy/astro-image-display-api/

    """

    def __init__(self, viewer):
        self.viewer = viewer
        self.app = viewer.jdaviz_app

    def _get_image_glue_data(self, image_label):
        if image_label is None:
            i_top = get_top_layer_index(self.viewer)
            image = self.viewer.layers[i_top].layer

        else:
            for lyr in self.viewer.layers:
                image = lyr.layer
                if image.label == image_label:
                    break
            else:
                raise ValueError(f"No data with data_label `{image_label}` found in viewer.")

        return image, image.label

    def _set_center(self, center):
        if center is None:
            return

        imviz_aligned_by_wcs = self.app._align_by == 'wcs'
        if isinstance(center, SkyCoord):
            if imviz_aligned_by_wcs:
                center = center.ra.degree, center.dec.degree
            else:
                reference_wcs = self.viewer.state.reference_data.coords

                if isinstance(reference_wcs, GWCS):
                    reference_wcs = WCS(reference_wcs.to_fits_sip())

                center = reference_wcs.world_to_pixel(center)

        elif not hasattr(center, '__len__') or not isinstance(center[0], (float, int)):
            raise ValueError(
                "The AID API supports `center` arguments as SkyCoords or as "
                f"a tuple of floats in pixel coordinates, got {center=}."
            )

        with delay_callback(self.viewer.state, "zoom_center_y", "zoom_center_x"):
            (
                self.viewer.state.zoom_center_x,
                self.viewer.state.zoom_center_y
            ) = center

    def _set_fov(self, fov):
        if fov is None:
            return

        if isinstance(fov, (u.Quantity, Angle)):
            current_fov = self._get_current_fov('sky')
        elif isinstance(fov, (float, int)):
            current_fov = self._get_current_fov('pixel')
        else:
            raise ValueError(
                f"`fov` must be a Quantity or tuple of floats, got {fov=}"
            )

        scale_factor = float(fov / current_fov)

        with delay_callback(self.viewer.state, "zoom_radius"):
            self.viewer.state.zoom_radius = self.viewer.state.zoom_radius * scale_factor

    def _set_rotation(self, rotation):
        if rotation is None:
            return

        orientation = self.app._jdaviz_helper.plugins.get('Orientation', None)

        if orientation.align_by != 'WCS':
            raise ValueError("The viewer must be aligned by WCS to use `set_rotation`.")

        if isinstance(rotation, (u.Quantity, Angle)):
            rotation = rotation.to_value(u.deg)
        elif not isinstance(rotation, (float, int)):
            raise ValueError(
                f"`rotation` must be a Quantity or float, got {rotation=}"
            )

        degn = orientation._obj._get_wcs_angles()[-3]
        rotation_angle = (degn + rotation) % 360

        label = f'{rotation:.2f} deg east of north'

        if label == orientation._obj.orientation.selected:
            return
        elif label in orientation._obj.orientation.choices:
            orientation._obj.orientation.selected = label
        else:
            orientation._obj.orientation.selected = "Default orientation"
            orientation.add_orientation(
                east_left=True,
                set_on_create=True,
                label=label,
                rotation_angle=rotation_angle
            )

    def set_viewport(self, center=None, fov=None, rotation=None, image_label=None, **kwargs):
        """
        Parameters
        ----------
        center : `~astropy.coordinates.SkyCoord` or tuple of floats
            Center the viewer on this coordinate.

        fov : `~astropy.units.Quantity` or tuple of floats
            Set the width of the viewport to span `field_of_view`.

            Set the viewport with respect to the image
            with the data label: ``image_label``.
        """
        with ignore_callback(
            self.viewer.state,
            'x_min', 'x_max', 'y_min', 'y_max',
            'zoom_center_x', 'zoom_center_y', 'zoom_radius'
        ):
            self._set_rotation(rotation)

        self._set_center(center)
        self._set_fov(fov)

    def _mean_pixel_scale(self, data):
        wcs = data.coords

        # for now, convert GWCS to FITS SIP so pixel to world
        # transformations can be done outside of the bounding box
        if isinstance(wcs, GWCS):
            wcs = WCS(wcs.to_fits_sip())

        abs_cdelts = u.Quantity([
            abs(cdelt) * u.Unit(cunit)
            for cdelt, cunit in zip(wcs.wcs.cdelt, wcs.wcs.cunit)
        ])
        return np.mean(abs_cdelts)

    def _get_current_fov(self, sky_or_pixel=None):
        state = self.viewer.state
        wcs = state.reference_data.coords

        pixel_fov = min(
            state.x_max - state.x_min,
            state.y_max - state.y_min
        )

        if sky_or_pixel == 'pixel':
            return pixel_fov
        if not any(c.strip() for c in wcs.wcs.ctype):
            raise ValueError("The image must have valid WCS to return `fov` in `sky`.")

        if isinstance(wcs, GWCS):
            wcs = WCS(wcs.to_fits_sip())

        # compute the mean of the height and width of the
        # viewer's FOV on ``data`` in world units:
        x_corners = [
            state.x_min,
            state.x_max,
            state.x_min
        ]
        y_corners = [
            state.y_min,
            state.y_min,
            state.y_max
        ]

        sky_corners = wcs.pixel_to_world(x_corners, y_corners)
        height_sky = sky_corners[0].separation(sky_corners[2])
        width_sky = sky_corners[0].separation(sky_corners[1])

        current_fov = min([width_sky, height_sky])

        return current_fov

    def _get_current_center(self, sky_or_pixel, image_label=None):
        # center pixel coordinates on the reference data:
        center_x = self.viewer.state.zoom_center_x
        center_y = self.viewer.state.zoom_center_y

        if self.app._align_by == 'wcs':
            reference_data = self.viewer.state.reference_data
        else:
            reference_data, image_label = self._get_image_glue_data(image_label)

        reference_wcs = reference_data.coords

        # # if the image data have WCS, get the center sky coordinate:
        if sky_or_pixel in ('sky', None):
            if self.app._align_by == 'wcs':
                center = self.viewer._get_center_skycoord()
            else:
                center = reference_wcs.pixel_to_world(center_x, center_y)
        else:
            center = (center_x, center_y)

        return center

    def _get_current_rotation(self):
        reference_data = self.viewer.state.reference_data
        degn = get_compass_info(
            reference_data.coords, reference_data.shape
        )[-3]
        rotation = Angle(-degn, unit=u.deg).wrap_at(360*u.deg)

        return rotation

    def get_viewport(self, sky_or_pixel=None, image_label=None, **kwargs):
        """
        sky_or_pixel : str, optional
            If 'sky', the center will be returned as a `SkyCoord` object.
            If 'pixel', the center will be returned as a tuple of pixel coordinates.
            If `None`, the default behavior is to return the center as a `SkyCoord` if
            possible, or as a tuple of floats if the image is in pixel coordinates and has
            no WCS information.

        image_label : str, optional
            The label of the image to get the viewport for. If not given and there is only one
            image loaded, the viewport for that image is returned. If there are multiple images
            and no label is provided, an error is raised.

        Returns
        -------
        dict
            A dictionary containing the current viewport settings.
            The keys are 'center', 'fov', and 'image_label'.
            - 'center' is an `astropy.coordinates.SkyCoord` object or a tuple of floats.
            - 'fov' is an `astropy.units.Quantity` object or a float.
            - 'image_label' is a string representing the label of the image.
        """
        image, image_label = self._get_image_glue_data(image_label)

        return dict(
            center=self._get_current_center(sky_or_pixel=sky_or_pixel, image_label=image_label),
            fov=self._get_current_fov(sky_or_pixel=sky_or_pixel),
            rotation=self._get_current_rotation(),
            image_label=image_label
        )



import pytest
from astropy import units as u
from astropy.coordinates import Angle, ICRS, SkyCoord
from astropy.tests.helper import assert_quantity_allclose
from glue.core.roi import CircularROI, EllipticalROI, RectangularROI
from numpy.testing import assert_allclose
from photutils.aperture import (CircularAperture, SkyCircularAperture,
                                EllipticalAperture, SkyEllipticalAperture,
                                RectangularAperture, SkyRectangularAperture,
                                CircularAnnulus, SkyCircularAnnulus,
                                EllipticalAnnulus, SkyEllipticalAnnulus,
                                RectangularAnnulus, SkyRectangularAnnulus)
from regions import (CirclePixelRegion, CircleSkyRegion,
                     EllipsePixelRegion, EllipseSkyRegion,
                     RectanglePixelRegion, CircleAnnulusPixelRegion,
                     EllipseAnnulusPixelRegion, RectangleAnnulusPixelRegion,
                     PolygonPixelRegion, PolygonSkyRegion,
                     PixCoord)

from jdaviz.core.region_translators import (
    _create_polygon_skyregion_from_coords, _create_circle_skyregion_from_coords,
    _create_ellipse_skyregion_from_coords, is_stcs_string, stcs_string2region,
    regions2roi, regions2aperture, aperture2regions)


# TODO: Use proper method from upstream when that is available.
def compare_region_shapes(reg1, reg2):
    assert reg1.__class__ == reg2.__class__
    for param in reg1._params:
        par1 = getattr(reg1, param)
        par2 = getattr(reg2, param)
        if isinstance(par1, PixCoord):
            assert_allclose(par1.xy, par2.xy)
        elif isinstance(par1, SkyCoord):
            assert par1 == par2
        elif isinstance(par1, u.Quantity):
            assert_quantity_allclose(par1, par2)
        else:
            assert_allclose(par1, par2)


def test_translation_circle(image_2d_wcs):
    region_shape = CirclePixelRegion(center=PixCoord(x=42, y=43), radius=4.2)
    aperture = regions2aperture(region_shape)
    assert isinstance(aperture, CircularAperture)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.r, region_shape.radius)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = regions2aperture(region_sky)
    assert isinstance(aperture_sky, SkyCircularAperture)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.r, region_sky.radius)

    # Roundtrip
    compare_region_shapes(region_shape, aperture2regions(aperture))
    compare_region_shapes(region_sky, aperture2regions(aperture_sky))

    with pytest.raises(ValueError, match='regions shape only accepts scalar x and y positions'):
        aperture2regions(CircularAperture(((10, 20), (30, 40)), 3))

    # NOTE: If these no longer fail, we also have to account for non-scalar inputs.
    # Assume this is representative for the sky counterpart too.
    with pytest.raises(ValueError, match='must be a scalar PixCoord'):
        CirclePixelRegion(center=PixCoord(x=[0, 42], y=[1, 43]), radius=4.2)
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        CirclePixelRegion(center=PixCoord(x=42, y=43), radius=[1, 4.2])


def test_translation_circle_roi(image_2d_wcs):
    region_shape = CirclePixelRegion(center=PixCoord(x=42, y=43), radius=4.2)
    roi_pix = regions2roi(region_shape)
    assert isinstance(roi_pix, CircularROI)
    assert_allclose((roi_pix.xc, roi_pix.yc, roi_pix.radius), (42, 43, 4.2))

    region_sky = region_shape.to_sky(image_2d_wcs)
    roi_sky = regions2roi(region_sky, wcs=image_2d_wcs)
    assert isinstance(roi_sky, CircularROI)
    assert_allclose((roi_sky.xc, roi_sky.yc, roi_sky.radius),
                    (roi_pix.xc, roi_pix.yc, roi_pix.radius))

    with pytest.raises(ValueError, match='WCS must be provided'):
        regions2roi(region_sky)


def test_translation_ellipse(image_2d_wcs):
    region_shape = EllipsePixelRegion(
        center=PixCoord(x=42, y=43), width=16, height=10, angle=Angle(30, 'deg'))
    aperture = regions2aperture(region_shape)
    assert isinstance(aperture, EllipticalAperture)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.a * 2, region_shape.width)
    assert_allclose(aperture.b * 2, region_shape.height)
    assert_quantity_allclose(aperture.theta, region_shape.angle)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = regions2aperture(region_sky)
    assert isinstance(aperture_sky, SkyEllipticalAperture)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.a * 2, region_sky.width)
    assert_quantity_allclose(aperture_sky.b * 2, region_sky.height)
    assert_quantity_allclose(aperture_sky.theta + (90 * u.deg), region_sky.angle)

    # Roundtrip
    compare_region_shapes(region_shape, aperture2regions(aperture))
    compare_region_shapes(region_sky, aperture2regions(aperture_sky))

    # NOTE: If these no longer fail, we also have to account for non-scalar inputs.
    # Assume this is representative for the sky counterpart too.
    with pytest.raises(ValueError, match='must be a scalar PixCoord'):
        EllipsePixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]), width=16, height=10, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipsePixelRegion(
            center=PixCoord(x=42, y=43), width=[1, 16], height=10, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipsePixelRegion(
            center=PixCoord(x=42, y=43), width=16, height=[1, 10], angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipsePixelRegion(
            center=PixCoord(x=42, y=43), width=16, height=10, angle=Angle([0, 30], 'deg'))


def test_translation_ellipse_roi(image_2d_wcs):
    region_shape = EllipsePixelRegion(
        center=PixCoord(x=42, y=43), width=16, height=10, angle=Angle(30, 'deg'))
    roi_pix = regions2roi(region_shape)
    assert isinstance(roi_pix, EllipticalROI)
    assert_allclose((roi_pix.xc, roi_pix.yc, roi_pix.radius_x, roi_pix.radius_y, roi_pix.theta),
                    (42, 43, 8.0, 5.0, 0.5235987755982988))

    region_sky = region_shape.to_sky(image_2d_wcs)
    roi_sky = regions2roi(region_sky, wcs=image_2d_wcs)
    assert isinstance(roi_sky, EllipticalROI)
    assert_allclose((roi_sky.xc, roi_sky.yc, roi_sky.radius_x, roi_sky.radius_y, roi_sky.theta),
                    (roi_pix.xc, roi_pix.yc, roi_pix.radius_x, roi_pix.radius_y, roi_pix.theta))


def test_translation_rectangle(image_2d_wcs):
    region_shape = RectanglePixelRegion(
        center=PixCoord(x=42, y=43), width=16, height=10, angle=Angle(30, 'deg'))
    aperture = regions2aperture(region_shape)
    assert isinstance(aperture, RectangularAperture)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.w, region_shape.width)
    assert_allclose(aperture.h, region_shape.height)
    assert_quantity_allclose(aperture.theta, region_shape.angle)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = regions2aperture(region_sky)
    assert isinstance(aperture_sky, SkyRectangularAperture)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.w, region_sky.width)
    assert_quantity_allclose(aperture_sky.h, region_sky.height)
    assert_quantity_allclose(aperture_sky.theta + (90 * u.deg), region_sky.angle)

    # Roundtrip
    compare_region_shapes(region_shape, aperture2regions(aperture))
    compare_region_shapes(region_sky, aperture2regions(aperture_sky))

    # NOTE: If these no longer fail, we also have to account for non-scalar inputs.
    # Assume this is representative for the sky counterpart too.
    with pytest.raises(ValueError, match='must be a scalar PixCoord'):
        RectanglePixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]), width=16, height=10, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectanglePixelRegion(
            center=PixCoord(x=42, y=43), width=[1, 16], height=10, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectanglePixelRegion(
            center=PixCoord(x=42, y=43), width=16, height=[1, 10], angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectanglePixelRegion(
            center=PixCoord(x=42, y=43), width=16, height=10, angle=Angle([0, 30], 'deg'))


def test_translation_rectangle_roi(image_2d_wcs):
    region_shape = RectanglePixelRegion(
        center=PixCoord(x=42, y=43), width=16, height=10, angle=Angle(30, 'deg'))
    roi_pix = regions2roi(region_shape)
    assert isinstance(roi_pix, RectangularROI)
    assert_allclose((roi_pix.xmin, roi_pix.xmax, roi_pix.ymin, roi_pix.ymax, roi_pix.theta),
                    (34, 50, 38, 48, 0.5235987755982988))

    region_sky = region_shape.to_sky(image_2d_wcs)
    roi_sky = regions2roi(region_sky, wcs=image_2d_wcs)
    assert isinstance(roi_sky, RectangularROI)
    assert_allclose((roi_sky.xmin, roi_sky.xmax, roi_sky.ymin, roi_sky.ymax, roi_sky.theta),
                    (roi_pix.xmin, roi_pix.xmax, roi_pix.ymin, roi_pix.ymax, roi_pix.theta))


def test_translation_circle_annulus(image_2d_wcs):
    region_shape = CircleAnnulusPixelRegion(
        center=PixCoord(x=42, y=43), inner_radius=5, outer_radius=8)
    aperture = regions2aperture(region_shape)
    assert isinstance(aperture, CircularAnnulus)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.r_in, region_shape.inner_radius)
    assert_allclose(aperture.r_out, region_shape.outer_radius)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = regions2aperture(region_sky)
    assert isinstance(aperture_sky, SkyCircularAnnulus)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.r_in, region_sky.inner_radius)
    assert_quantity_allclose(aperture_sky.r_out, region_sky.outer_radius)

    # Roundtrip
    compare_region_shapes(region_shape, aperture2regions(aperture))
    compare_region_shapes(region_sky, aperture2regions(aperture_sky))

    # NOTE: If these no longer fail, we also have to account for non-scalar inputs.
    # Assume this is representative for the sky counterpart too.
    with pytest.raises(ValueError, match='must be a scalar PixCoord'):
        CircleAnnulusPixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]), inner_radius=5, outer_radius=8)
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        CircleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_radius=[1, 5], outer_radius=8)
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        CircleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_radius=5, outer_radius=[8, 10])


def test_translation_ellipse_annulus(image_2d_wcs):
    region_shape = EllipseAnnulusPixelRegion(
        center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=8.5,
        outer_height=6.5, angle=Angle(30, 'deg'))
    aperture = regions2aperture(region_shape)
    assert isinstance(aperture, EllipticalAnnulus)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.a_in * 2, region_shape.inner_width)
    assert_allclose(aperture.a_out * 2, region_shape.outer_width)
    assert_allclose(aperture.b_in * 2, region_shape.inner_height)
    assert_allclose(aperture.b_out * 2, region_shape.outer_height)
    assert_quantity_allclose(aperture.theta, region_shape.angle)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = regions2aperture(region_sky)
    assert isinstance(aperture_sky, SkyEllipticalAnnulus)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.a_in * 2, region_sky.inner_width)
    assert_quantity_allclose(aperture_sky.a_out * 2, region_sky.outer_width)
    assert_quantity_allclose(aperture_sky.b_in * 2, region_sky.inner_height)
    assert_quantity_allclose(aperture_sky.b_out * 2, region_sky.outer_height)
    assert_quantity_allclose(aperture_sky.theta + (90 * u.deg), region_sky.angle)

    # Roundtrip
    compare_region_shapes(region_shape, aperture2regions(aperture))
    compare_region_shapes(region_sky, aperture2regions(aperture_sky))

    # NOTE: If these no longer fail, we also have to account for non-scalar inputs.
    # Assume this is representative for the sky counterpart too.
    with pytest.raises(ValueError, match='must be a scalar PixCoord'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]), inner_width=5.5, inner_height=3.5,
            outer_width=8.5, outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=[1, 5.5], inner_height=3.5, outer_width=8.5,
            outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=[1, 3.5], outer_width=8.5,
            outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=[8.5, 10],
            outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=8.5,
            outer_height=[6.5, 10], angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=8.5,
            outer_height=6.5, angle=Angle([0, 30], 'deg'))


def test_translation_rectangle_annulus(image_2d_wcs):
    region_shape = RectangleAnnulusPixelRegion(
        center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=8.5,
        outer_height=6.5, angle=Angle(30, 'deg'))
    aperture = regions2aperture(region_shape)
    assert isinstance(aperture, RectangularAnnulus)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.w_in, region_shape.inner_width)
    assert_allclose(aperture.w_out, region_shape.outer_width)
    assert_allclose(aperture.h_in, region_shape.inner_height)
    assert_allclose(aperture.h_out, region_shape.outer_height)
    assert_quantity_allclose(aperture.theta, region_shape.angle)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = regions2aperture(region_sky)
    assert isinstance(aperture_sky, SkyRectangularAnnulus)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.w_in, region_sky.inner_width)
    assert_quantity_allclose(aperture_sky.w_out, region_sky.outer_width)
    assert_quantity_allclose(aperture_sky.h_in, region_sky.inner_height)
    assert_quantity_allclose(aperture_sky.h_out, region_sky.outer_height)
    assert_quantity_allclose(aperture_sky.theta + (90 * u.deg), region_sky.angle)

    # Roundtrip
    compare_region_shapes(region_shape, aperture2regions(aperture))
    compare_region_shapes(region_sky, aperture2regions(aperture_sky))

    # NOTE: If these no longer fail, we also have to account for non-scalar inputs.
    # Assume this is representative for the sky counterpart too.
    with pytest.raises(ValueError, match='must be a scalar PixCoord'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]), inner_width=5.5, inner_height=3.5,
            outer_width=8.5, outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=[1, 5.5], inner_height=3.5, outer_width=8.5,
            outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=[1, 3.5], outer_width=8.5,
            outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=[8.5, 10],
            outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=8.5,
            outer_height=[6.5, 10], angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=8.5,
            outer_height=6.5, angle=Angle([0, 30], 'deg'))


def test_translation_polygon():
    region_shape = PolygonPixelRegion(vertices=PixCoord(x=[1, 2, 2], y=[1, 1, 2]))
    with pytest.raises(NotImplementedError, match='is not supported'):
        regions2aperture(region_shape)
    with pytest.raises(NotImplementedError, match='is not supported'):
        regions2roi(region_shape)


def test___create_polygon_skyregion_from_coords():
    coords = [10.0, 20.0, 30.0, 40.0, 50.0, 60.0]
    reg = _create_polygon_skyregion_from_coords(coords)
    assert isinstance(reg, PolygonSkyRegion)
    assert_allclose(reg.vertices.ra.degree, [10, 30, 50])
    assert_allclose(reg.vertices.dec.degree, [20, 40, 60])
    # Check that the default frame is ICRS
    assert isinstance(reg.vertices.frame, ICRS)
    # Check that the default unit is deg
    assert reg.vertices.ra.unit == u.deg


def test___create_circle_skyregion_from_coords():
    coords = [10.0, 20.0, 30.0]
    reg = _create_circle_skyregion_from_coords(coords)
    assert isinstance(reg, CircleSkyRegion)
    assert_allclose(reg.center.ra.degree, 10)
    assert_allclose(reg.center.dec.degree, 20)
    # Check that the default frame is ICRS
    assert isinstance(reg.center.frame, ICRS)
    # Check that the default unit is deg
    assert reg.center.ra.unit == u.deg


def test___create_ellipse_skyregion_from_coords():
    coords = [10.0, 20.0, 30.0, 40.0, 50.0]
    reg = _create_ellipse_skyregion_from_coords(coords)
    assert isinstance(reg, EllipseSkyRegion)
    assert_allclose(reg.center.ra.degree, 10)
    assert_allclose(reg.center.dec.degree, 20)
    # Check that the default frame is ICRS
    assert isinstance(reg.center.frame, ICRS)
    # Check that the default unit is deg
    assert reg.center.ra.unit == u.deg


@pytest.mark.parametrize('stcs_string, expected', [
    ('POLYGON ICRS 10.0 20.0', True),
    ('CIRCLE ICRS 10.0 20.0', True),
    ('ELLIPSE ICRS 10.0 20.0', True),
    ('polygon icrs 10.0 20.0', True),
    ('circle icrs 10.0 20.0', True),
    ('ellipse icrs 10.0 20.0', True),
    ('POLYGON J2000 10.0 20.0', True),
    ('CIRCLE J2000 10.0 20.0', True),
    ('ELLIPSE J2000 10.0 20.0', True),
    ('POLYGON FK5 10.0 20.0', True),
    ('CIRCLE FK5 10.0 20.0', True),
    ('ELLIPSE FK5 10.0 20.0', True),
    ('POLYGON 10.0 20.0', True),
    ('CIRCLE 10.0 20.0', True),
    ('ELLIPSE 10.0 20.0', True),
    ('POLYGON ICRS 1.0', False),
    ('POLYGON ICRS 1 2 3', False),
    ('POLYGON 1.0', False),
    ('POLYGON 1 2 3', False),
    ('not a STC-S string', False),
    # Invalid input type (not a string) should return False instead of raising an error
    (123, False),
])
def test_is_stcs_string(stcs_string, expected):
    assert is_stcs_string(stcs_string) is expected


@pytest.mark.parametrize('stcs_string, expected', [
    ('POLYGON ICRS 10.0 20.0 30.0 40.0 50.0 60.0',
     PolygonSkyRegion(vertices=SkyCoord(ra=[10, 30, 50], dec=[20, 40, 60], unit='deg'))),
    ('CIRCLE ICRS 10.0 20.0 30.0',
     CircleSkyRegion(center=SkyCoord(ra=10, dec=20, unit='deg'), radius=30 * u.deg)),
    ('ELLIPSE ICRS 10.0 20.0 30.0 40.0 50.0',
     EllipseSkyRegion(center=SkyCoord(ra=10, dec=20, unit='deg'),
                      width=30 * u.deg, height=40 * u.deg, angle=50 * u.deg)),
])
def test_stcs_string2region_correctness(stcs_string, expected):
    reg = stcs_string2region(stcs_string)
    assert reg == expected


@pytest.mark.parametrize('stcs_string, exception_message', [
    ('CIRCLE ICRS 10', 'Invalid STC-S string'),
    (123, 'STC-S string must be a string.'),
])
def test_stcs_string2region_exceptions(stcs_string, exception_message):
    with pytest.raises(ValueError, match=exception_message):
        stcs_string2region(stcs_string)



from functools import cached_property
from astropy.io import fits

from jdaviz.core.loaders.parsers import BaseParser
from jdaviz.core.registries import loader_parser_registry


__all__ = ['FITSParser']


@loader_parser_registry('fits')
class FITSParser(BaseParser):

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'specviz2d',
                                   'lcviz', 'imviz', 'cubeviz',
                                   'rampviz'):
            # NOTE: temporary during deconfig process
            return False

        try:
            self.output
        except Exception:
            return False

        return True

    @cached_property
    def output(self):
        return fits.open(self.input)

    def _cleanup(self):
        if 'output' not in self.__dict__:
            return
        for hdu in self.output:
            try:
                del hdu.data
            except Exception:  # nosec
                pass
        try:
            self.output.close()
        except Exception:  # nosec
            pass
        self._clear_cache('output')
